import platform
import textwrap

import pytest

from conans.model.ref import ConanFileReference
from conans.test.assets.genconanfile import GenConanfile
from conans.test.assets.sources import gen_function_h, gen_function_cpp
from conans.test.utils.tools import TestClient


@pytest.mark.slow
@pytest.mark.tool_cmake
@pytest.fixture(scope="module")
def setup_client_with_greetings():
    """
    creates a multi-component package with 2 components "hello" and "bye
    """
    hello_h = gen_function_h(name="hello")
    hello_cpp = gen_function_cpp(name="hello", includes=["hello"])
    bye_h = gen_function_h(name="bye")
    bye_cpp = gen_function_cpp(name="bye", includes=["bye"])

    conanfile_greetings = textwrap.dedent("""
        import os
        from conans import ConanFile, CMake, tools

        class GreetingsConan(ConanFile):
            name = "greetings"
            version = "0.0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "cmake"
            exports_sources = "src/*"
            options = {"components": ["standard", "custom", "none"]}
            default_options = {"components": "standard"}

            def build(self):
                cmake = CMake(self)
                cmake.configure(source_folder="src")
                cmake.build()

            def package(self):
                self.copy("*.h", dst="include", src="src")
                self.copy("*.lib", dst="lib", keep_path=False)
                self.copy("*.a", dst="lib", keep_path=False)
                tools.save(os.path.join(self.package_folder,
                                        "my_cmake_functions.cmake"), "set(MY_CMAKE_VAR 99)")

            def package_info(self):

                if self.options.components == "standard":
                    self.cpp_info.components["hello"].libs = ["hello"]
                    self.cpp_info.components["bye"].libs = ["bye"]
                    self.cpp_info.components["hello"].build_modules = ["my_cmake_functions.cmake"]
                elif self.options.components == "custom":
                    self.cpp_info.filenames["cmake_find_package_multi"] = "MYG"
                    self.cpp_info.filenames["cmake_find_package"] = "MYG"
                    self.cpp_info.set_property("cmake_file_name", "MYG")

                    self.cpp_info.names["cmake_find_package_multi"] = "MyGreetings"
                    self.cpp_info.names["cmake_find_package"] = "MyGreetings"
                    self.cpp_info.set_property("cmake_target_name", "MyGreetings")

                    self.cpp_info.components["hello"].names["cmake_find_package_multi"] = "MyHello"
                    self.cpp_info.components["bye"].names["cmake_find_package_multi"] = "MyBye"
                    self.cpp_info.components["hello"].names["cmake_find_package"] = "MyHello"
                    self.cpp_info.components["bye"].names["cmake_find_package"] = "MyBye"
                    self.cpp_info.components["hello"].set_property("cmake_target_name", "MyHello")
                    self.cpp_info.components["bye"].set_property("cmake_target_name", "MyBye")

                    self.cpp_info.components["hello"].libs = ["hello"]
                    self.cpp_info.components["bye"].libs = ["bye"]

                    # Duplicated on purpose to check that it doesn't break
                    self.cpp_info.components["bye"].build_modules = ["my_cmake_functions.cmake"]
                    self.cpp_info.components["hello"].build_modules = ["my_cmake_functions.cmake"]

                else:
                    self.cpp_info.libs = ["hello", "bye"]

            def package_id(self):
                del self.info.options.components
        """)

    cmakelists_greetings = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(greetings CXX)

        include(${CMAKE_BINARY_DIR}/conanbuildinfo.cmake)
        conan_output_dirs_setup()

        add_library(hello hello.cpp)
        add_library(bye bye.cpp)
        """)

    test_package_greetings_conanfile = textwrap.dedent("""
        import os
        from conans import ConanFile, CMake

        class GreetingsTestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "cmake", "cmake_find_package_multi"
            requires = "greetings/0.0.1"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def test(self):
                os.chdir("bin")
                self.run(".%sexample" % os.sep)
        """)
    test_package_greetings_cpp = gen_function_cpp(name="main", includes=["hello", "bye"],
                                                  calls=["hello", "bye"])

    test_package_greetings_cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(PackageTest CXX)

        include(${CMAKE_BINARY_DIR}/conanbuildinfo.cmake)
        conan_output_dirs_setup()

        find_package(greetings)

        add_executable(example example.cpp)
        target_link_libraries(example greetings::greetings)
        """)

    client = TestClient()
    client.save({"conanfile.py": conanfile_greetings,
                 "src/CMakeLists.txt": cmakelists_greetings,
                 "src/hello.h": hello_h,
                 "src/hello.cpp": hello_cpp,
                 "src/bye.h": bye_h,
                 "src/bye.cpp": bye_cpp,
                 "test_package/conanfile.py": test_package_greetings_conanfile,
                 "test_package/example.cpp": test_package_greetings_cpp,
                 "test_package/CMakeLists.txt": test_package_greetings_cmakelists})
    client.run("create . -s build_type=Release")
    assert "hello: Release!" in client.out
    assert "bye: Release!" in client.out
    client.run("create . -s build_type=Debug")
    assert "hello: Debug!" in client.out
    assert "bye: Debug!" in client.out
    return client


def create_chat(client, generator, components, package_info, cmake_find, test_cmake_find):
    conanfile = textwrap.dedent("""
        from conans import ConanFile, CMake

        class Chat(ConanFile):
            name = "chat"
            version = "0.0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "{}", "cmake"
            exports_sources = "src/*"
            requires = "greetings/0.0.1"
            default_options = {{"greetings/*:components": "{}"}}

            def build(self):
                cmake = CMake(self)
                cmake.configure(source_folder="src")
                cmake.build()

            def package(self):
                self.copy("*.h", dst="include", src="src")
                self.copy("*.lib", dst="lib", keep_path=False)
                self.copy("*.a", dst="lib", keep_path=False)

            def package_info(self):
                {}
        """).format(generator, components, "\n        ".join(package_info.splitlines()))
    sayhello_h = gen_function_h(name="sayhello")
    sayhello_cpp = gen_function_cpp(name="sayhello", includes=["hello"], calls=["hello"])
    sayhellobye_h = gen_function_h(name="sayhellobye")
    sayhellobye_cpp = gen_function_cpp(name="sayhellobye", includes=["sayhello", "bye"],
                                       calls=["sayhello", "bye"])

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(world CXX)

        include(${CMAKE_BINARY_DIR}/conanbuildinfo.cmake)
        conan_output_dirs_setup()

        %s
        """ % cmake_find)

    test_conanfile = textwrap.dedent("""
        import os
        from conans import ConanFile, CMake

        class WorldTestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "cmake", "{}"
            requires = "chat/0.0.1"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def test(self):
                os.chdir("bin")
                self.run(".%sexample" % os.sep)
                self.run(".%sexample2" % os.sep)
        """.format(generator))
    test_example_cpp = gen_function_cpp(name="main", includes=["sayhellobye"], calls=["sayhellobye"])

    test_cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(PackageTest CXX)

        include(${CMAKE_BINARY_DIR}/conanbuildinfo.cmake)
        conan_output_dirs_setup()

        # necessary for the local
        set(CMAKE_MODULE_PATH ${CMAKE_BINARY_DIR} ${CMAKE_MODULE_PATH})
        set(CMAKE_PREFIX_PATH ${CMAKE_BINARY_DIR} ${CMAKE_PREFIX_PATH})
        %s
        """ % test_cmake_find)
    client.save({"conanfile.py": conanfile,
                 "src/CMakeLists.txt": cmakelists,
                 "src/sayhello.h": sayhello_h,
                 "src/sayhello.cpp": sayhello_cpp,
                 "src/sayhellobye.h": sayhellobye_h,
                 "src/sayhellobye.cpp": sayhellobye_cpp,
                 "test_package/conanfile.py": test_conanfile,
                 "test_package/CMakeLists.txt": test_cmakelists,
                 "test_package/example.cpp": test_example_cpp}, clean_first=True)
    client.run("create . -s build_type=Release")
    assert "sayhellobye: Release!" in client.out
    assert "sayhello: Release!" in client.out
    assert "hello: Release!" in client.out
    assert "bye: Release!" in client.out
    client.run("create . -s build_type=Debug")
    assert "sayhellobye: Debug!" in client.out
    assert "sayhello: Debug!" in client.out
    assert "hello: Debug!" in client.out
    assert "bye: Debug!" in client.out


@pytest.mark.parametrize("generator", ["cmake_find_package_multi", "cmake_find_package"])
def test_standard_names(setup_client_with_greetings, generator):
    client = setup_client_with_greetings

    package_info = textwrap.dedent("""
        self.cpp_info.components["sayhello"].requires = ["greetings::hello"]
        self.cpp_info.components["sayhello"].libs = ["sayhello"]
        self.cpp_info.components["sayhellobye"].requires = ["sayhello", "greetings::bye"]
        self.cpp_info.components["sayhellobye"].libs = ["sayhellobye"]
        """)
    cmake_find = textwrap.dedent("""
        find_package(greetings COMPONENTS hello bye)

        add_library(sayhello sayhello.cpp)
        target_link_libraries(sayhello greetings::hello)

        add_library(sayhellobye sayhellobye.cpp)
        target_link_libraries(sayhellobye sayhello greetings::bye)
        """)
    test_cmake_find = textwrap.dedent("""
        find_package(chat)

        add_executable(example example.cpp)
        target_link_libraries(example chat::sayhellobye)

        add_executable(example2 example.cpp)
        target_link_libraries(example2 chat::chat)
        """)

    create_chat(client, generator, "standard", package_info, cmake_find, test_cmake_find)

    # Test consumer multi-config
    if generator == "cmake_find_package_multi" and platform.system() == "Windows":
        with client.chdir("test_package"):
            client.run("install . -s build_type=Release")
            client.run("install . -s build_type=Debug")
            client.run_command('cmake . -G "Visual Studio 15 Win64"')
            client.run_command("cmake --build . --config Debug")
            client.run_command(r".\bin\example.exe")
            assert "sayhellobye: Debug!" in client.out
            assert "sayhello: Debug!" in client.out
            assert "hello: Debug!" in client.out
            assert "bye: Debug!" in client.out
            client.run_command("cmake --build . --config Release")
            client.run_command(r".\bin\example.exe")
            assert "sayhellobye: Release!" in client.out
            assert "sayhello: Release!" in client.out
            assert "hello: Release!" in client.out
            assert "bye: Release!" in client.out


@pytest.mark.parametrize("generator", ["cmake_find_package_multi", "cmake_find_package"])
def test_custom_names(setup_client_with_greetings, generator):
    client = setup_client_with_greetings

    package_info = textwrap.dedent("""
        self.cpp_info.names["cmake_find_package_multi"] = "MyChat"
        self.cpp_info.names["cmake_find_package"] = "MyChat"
        # NOTE: For the new CMakeDeps only filenames mean filename, it is not using the "names" field
        self.cpp_info.set_property("cmake_target_name", "MyChat")
        self.cpp_info.set_property("cmake_file_name", "MyChat")

        self.cpp_info.components["sayhello"].names["cmake_find_package_multi"] = "MySay"
        self.cpp_info.components["sayhello"].names["cmake_find_package"] = "MySay"
        self.cpp_info.components["sayhello"].set_property("cmake_target_name", "MySay")

        self.cpp_info.components["sayhello"].requires = ["greetings::hello"]
        self.cpp_info.components["sayhello"].libs = ["sayhello"]
        self.cpp_info.components["sayhellobye"].names["cmake_find_package_multi"] ="MySayBye"
        self.cpp_info.components["sayhellobye"].names["cmake_find_package"] ="MySayBye"
        self.cpp_info.components["sayhellobye"].set_property("cmake_target_name", "MySayBye")

        self.cpp_info.components["sayhellobye"].requires = ["sayhello", "greetings::bye"]
        self.cpp_info.components["sayhellobye"].libs = ["sayhellobye"]
        """)

    cmake_find = textwrap.dedent("""
        find_package(MYG COMPONENTS MyHello MyBye)

        add_library(sayhello sayhello.cpp)
        target_link_libraries(sayhello MyGreetings::MyHello)

        add_library(sayhellobye sayhellobye.cpp)
        target_link_libraries(sayhellobye sayhello MyGreetings::MyBye)
        """)

    test_cmake_find = textwrap.dedent("""
        find_package(MyChat)

        add_executable(example example.cpp)
        target_link_libraries(example MyChat::MySayBye)

        add_executable(example2 example.cpp)
        target_link_libraries(example2 MyChat::MyChat)
        """)
    create_chat(client, generator, "custom", package_info, cmake_find, test_cmake_find)


@pytest.mark.parametrize("generator", ["cmake_find_package_multi", "cmake_find_package"])
def test_no_components(setup_client_with_greetings, generator):
    client = setup_client_with_greetings

    package_info = textwrap.dedent("""
        self.cpp_info.components["sayhello"].requires = ["greetings::greetings"]
        self.cpp_info.components["sayhello"].libs = ["sayhello"]
        self.cpp_info.components["sayhellobye"].requires = ["sayhello", "greetings::greetings"]
        self.cpp_info.components["sayhellobye"].libs = ["sayhellobye"]
        """)

    cmake_find = textwrap.dedent("""
        find_package(greetings)

        add_library(sayhello sayhello.cpp)
        target_link_libraries(sayhello greetings::greetings)

        add_library(sayhellobye sayhellobye.cpp)
        target_link_libraries(sayhellobye sayhello greetings::greetings)
        """)

    test_cmake_find = textwrap.dedent("""
        find_package(chat)

        add_executable(example example.cpp)
        target_link_libraries(example chat::sayhellobye)

        add_executable(example2 example.cpp)
        target_link_libraries(example2 chat::chat)
        """)
    create_chat(client, generator, "none", package_info, cmake_find, test_cmake_find)


@pytest.mark.parametrize("generator", ["cmake_find_package_multi", "cmake_find_package"])
@pytest.mark.slow
@pytest.mark.tool_cmake
def test_same_names(generator):
    client = TestClient()
    conanfile_greetings = textwrap.dedent("""
        from conans import ConanFile, CMake

        class HelloConan(ConanFile):
            name = "hello"
            version = "0.0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "cmake"
            exports_sources = "src/*"

            def build(self):
                cmake = CMake(self)
                cmake.configure(source_folder="src")
                cmake.build()

            def package(self):
                self.copy("*.h", dst="include", src="src")
                self.copy("*.lib", dst="lib", keep_path=False)
                self.copy("*.a", dst="lib", keep_path=False)

            def package_info(self):
                self.cpp_info.components["global"].name = "hello"
                self.cpp_info.components["global"].libs = ["hello"]
        """)
    hello_h = gen_function_h(name="hello")
    hello_cpp = gen_function_cpp(name="hello", includes=["hello"])

    cmakelists_greetings = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(greetings CXX)

        include(${CMAKE_BINARY_DIR}/conanbuildinfo.cmake)
        conan_basic_setup()

        add_library(hello hello.cpp)
        """)
    test_package_greetings_conanfile = textwrap.dedent("""
        import os
        from conans import ConanFile, CMake

        class HelloTestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "cmake", "{}"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def test(self):
                os.chdir("bin")
                self.run(".%sexample" % os.sep)
        """.format(generator))
    test_package_greetings_cpp = gen_function_cpp(name="main", includes=["hello"], calls=["hello"])

    test_package_greetings_cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(PackageTest CXX)

        include(${CMAKE_BINARY_DIR}/conanbuildinfo.cmake)
        conan_basic_setup()

        find_package(hello)

        add_executable(example example.cpp)
        target_link_libraries(example hello::hello)
        """)
    client.save({"conanfile.py": conanfile_greetings,
                 "src/CMakeLists.txt": cmakelists_greetings,
                 "src/hello.h": hello_h,
                 "src/hello.cpp": hello_cpp,
                 "test_package/conanfile.py": test_package_greetings_conanfile,
                 "test_package/example.cpp": test_package_greetings_cpp,
                 "test_package/CMakeLists.txt": test_package_greetings_cmakelists})
    client.run("create .")
    assert "hello: Release!" in client.out


@pytest.mark.tool_cmake
class TestComponentsCMakeGenerators:

    @pytest.mark.parametrize("generator", ["cmake_find_package_multi", "cmake_find_package"])
    def test_component_not_found(self, generator):
        conanfile = textwrap.dedent("""
            from conans import ConanFile

            class GreetingsConan(ConanFile):
                def package_info(self):
                    self.cpp_info.components["hello"].libs = ["hello"]
        """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . greetings/0.0.1@")

        conanfile = textwrap.dedent("""
            from conans import ConanFile

            class WorldConan(ConanFile):
                requires = "greetings/0.0.1"

                def package_info(self):
                    self.cpp_info.components["helloworld"].requires = ["greetings::non-existent"]
                    self.cpp_info.components["helloworld"].libs = ["helloworld"]
        """)
        client.save({"conanfile.py": conanfile})
        client.run("create . world/0.0.1@")
        client.run("install world/0.0.1@ -g {}".format(generator), assert_error=True)
        assert ("Component 'greetings::non-existent' not found in 'greetings' "
                "package requirement" in client.out)

    @pytest.mark.parametrize("generator, filename", [("cmake_find_package_multi", "greetings"),
                                                     ("cmake_find_package_multi", "filegreetings"),
                                                     ("cmake_find_package", "greetings"),
                                                     ("cmake_find_package", "filegreetings")])
    def test_component_not_found_cmake(self, generator, filename):
        conanfile = textwrap.dedent("""
            from conans import ConanFile

            class GreetingsConan(ConanFile):
                def package_info(self):
                    self.cpp_info.components["hello"].libs = ["hello"]
                    self.cpp_info.filenames["{}"] = "{}"
        """.format(generator, filename))
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . greetings/0.0.1@")

        conanfile = textwrap.dedent("""
            from conans import ConanFile, CMake

            class ConsumerConan(ConanFile):
                settings = "build_type"
                generators = "{}"
                requires = "greetings/0.0.1"

                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
        """.format(generator))
        cmakelists = textwrap.dedent("""
            set(CMAKE_CXX_COMPILER_WORKS 1)
            set(CMAKE_CXX_ABI_COMPILED 1)
            cmake_minimum_required(VERSION 3.0)
            project(Consumer CXX)

            find_package({filename} COMPONENTS hello)
            find_package({filename} COMPONENTS non-existent)
            """.format(filename=filename))
        client.save({"conanfile.py": conanfile, "CMakeLists.txt": cmakelists})
        client.run("install .")
        client.run("build .", assert_error=True)
        assert "Conan: Component 'hello' found in package 'greetings'" in client.out
        assert "Conan: Component 'non-existent' NOT found in package 'greetings'" in client.out

    @pytest.mark.parametrize("generator", ["cmake_find_package_multi", "cmake_find_package"])
    def test_component_not_found_same_name_as_pkg_require(self, generator):
        zlib = GenConanfile("zlib", "0.1").with_setting("build_type").with_generator(generator)
        mypkg = GenConanfile("mypkg", "0.1").with_setting("build_type").with_generator(generator)
        final = GenConanfile("final", "0.1").with_setting("build_type").with_generator(generator)\
            .with_require(ConanFileReference("zlib", "0.1", None, None))\
            .with_require(ConanFileReference("mypkg", "0.1", None, None))\
            .with_package_info(cpp_info={"components": {"cmp": {"requires": ["mypkg::zlib",
                                                                             "zlib::zlib"]}}},
                               env_info={})
        consumer = GenConanfile("consumer", "0.1").with_setting("build_type")\
            .with_generator(generator)\
            .with_requirement(ConanFileReference("final", "0.1", None, None))
        client = TestClient()
        client.save({"zlib.py": zlib, "mypkg.py": mypkg, "final.py": final, "consumer.py": consumer})
        client.run("create zlib.py")
        client.run("create mypkg.py")
        client.run("create final.py")
        client.run("install consumer.py", assert_error=True)
        assert "Component 'mypkg::zlib' not found in 'mypkg' package requirement" in client.out

    @pytest.mark.slow
    @pytest.mark.parametrize("generator", ["cmake_find_package_multi", "cmake_find_package"])
    def test_same_name_global_target_collision(self, generator):
        # https://github.com/conan-io/conan/issues/7889
        conanfile_tpl = textwrap.dedent("""
            from conans import ConanFile, CMake

            class Conan(ConanFile):
                name = "{name}"
                version = "1.0"
                settings = "os", "compiler", "build_type", "arch"
                generators = "cmake"
                exports_sources = "src/*"

                def build(self):
                    cmake = CMake(self)
                    cmake.configure(source_folder="src")
                    cmake.build()

                def package(self):
                    self.copy("*.h", dst="include", src="src")
                    self.copy("*.lib", dst="lib", keep_path=False)
                    self.copy("*.a", dst="lib", keep_path=False)

                def package_info(self):
                    self.cpp_info.names["{generator}"] = "nonstd"
                    self.cpp_info.filenames["{generator}"] = "{name}"
                    self.cpp_info.set_property("cmake_target_name", "nonstd")
                    self.cpp_info.set_property("cmake_file_name", "{name}")

                    self.cpp_info.components["1"].names["{generator}"] = "{name}"
                    self.cpp_info.components["1"].set_property("cmake_target_name", "{name}")
                    self.cpp_info.components["1"].libs = ["{name}"]
            """)
        client = TestClient()
        for name in ["expected", "variant"]:
            client.run("new {name}/1.0 -s".format(name=name))
            client.save({"conanfile.py": conanfile_tpl.format(name=name, generator=generator)})
            client.run("create . {name}/1.0@".format(name=name))
        middle_cmakelists = textwrap.dedent("""
            set(CMAKE_CXX_COMPILER_WORKS 1)
            set(CMAKE_CXX_ABI_COMPILED 1)
            project(middle CXX)
            cmake_minimum_required(VERSION 3.1)

            include(${CMAKE_BINARY_DIR}/conanbuildinfo.cmake)
            conan_basic_setup()

            find_package(expected)
            find_package(variant)

            add_library(middle middle.cpp)
            target_link_libraries(middle nonstd::nonstd)
            """)
        middle_h = gen_function_h(name="middle")
        middle_cpp = gen_function_cpp(name="middle", includes=["middle", "expected", "variant"],
                                      calls=["expected", "variant"])
        middle_conanfile = textwrap.dedent("""
            from conans import ConanFile, CMake

            class Conan(ConanFile):
                name = "middle"
                version = "1.0"
                settings = "os", "compiler", "build_type", "arch"
                generators = "cmake", "{}"
                exports_sources = "src/*"
                requires = "expected/1.0", "variant/1.0"

                def build(self):
                    cmake = CMake(self)
                    cmake.configure(source_folder="src")
                    cmake.build()

                def package(self):
                    self.copy("*.h", dst="include", src="src")
                    self.copy("*.lib", dst="lib", keep_path=False)
                    self.copy("*.a", dst="lib", keep_path=False)

                def package_info(self):
                    self.cpp_info.libs = ["middle"]
            """.format(generator))
        client.save({"conanfile.py": middle_conanfile, "src/CMakeLists.txt": middle_cmakelists,
                     "src/middle.h": middle_h, "src/middle.cpp": middle_cpp}, clean_first=True)
        client.run("create . middle/1.0@")
        conanfile = textwrap.dedent("""
            import os
            from conans import ConanFile, CMake

            class Conan(ConanFile):
                name = "consumer"
                version = "1.0"
                generators = "{}", "cmake"
                settings = "os", "compiler", "build_type", "arch"
                exports_sources = "src/*"
                requires = "middle/1.0"

                def build(self):
                    cmake = CMake(self)
                    cmake.configure(source_folder="src")
                    cmake.build()
                    self.run(os.path.join("bin", "main"))
            """.format(generator))
        cmakelists = textwrap.dedent("""
            set(CMAKE_CXX_COMPILER_WORKS 1)
            set(CMAKE_CXX_ABI_COMPILED 1)
            project(consumer CXX)
            cmake_minimum_required(VERSION 3.1)

            include(${CMAKE_BINARY_DIR}/conanbuildinfo.cmake)
            conan_basic_setup()

            find_package(middle)

            get_target_property(tmp middle::middle INTERFACE_LINK_LIBRARIES)
            message("Middle link libraries: ${tmp}")

            add_executable(main main.cpp)
            target_link_libraries(main middle::middle)
            """)
        main_cpp = gen_function_cpp(name="main", includes=["middle"], calls=["middle"])
        client.save({"conanfile.py": conanfile,
                     "src/CMakeLists.txt": cmakelists,
                     "src/main.cpp": main_cpp}, clean_first=True)
        client.run("create . consumer/1.0@")
        # assert False
        assert 'main: Release!' in client.out
        assert 'middle: Release!' in client.out
        assert 'expected/1.0: Hello World Release!' in client.out
        assert 'variant/1.0: Hello World Release!' in client.out



import os
import re
import unittest

from six.moves.urllib.parse import unquote

from conans import MATRIX_PARAMS
from conans.test.assets.genconanfile import GenConanfile
from conans.test.utils.tools import TestClient, TestRequester, TestServer
from conans.util.files import save


class PutPropertiesTest(unittest.TestCase):

    def test_create_empty_property_file(self):
        client = TestClient()
        client.save({"conanfile.py": GenConanfile("Hello", "0.1")})
        client.run("export . lasote/stable")
        props_file = client.cache.artifacts_properties_path
        self.assertTrue(os.path.exists(props_file))

    def test_put_properties(self):
        test_server = TestServer()
        servers = {"default": test_server}

        wanted_vars = {"key0": "value",
                       "key1": "with space",
                       "key2": "with/slash",
                       "key3": "with.dot",
                       "key4": "with;semicolon",
                       "key5": "with~virgul",
                       "key6": "with#hash"
                       }

        class RequesterCheckArtifactProperties(TestRequester):
            def put(self_requester, url, **kwargs):
                # Check headers
                for name, value in wanted_vars.items():
                    value1 = kwargs["headers"][name]
                    self.assertEqual(value1, value)

                # Check matrix params
                self.assertNotIn(';', url)
                mp = re.match(r"^[^;\s]+;(?P<matrix_params>[^/]+)/.*", url)
                self.assertFalse(mp)

                return super(RequesterCheckArtifactProperties, self_requester).put(url, **kwargs)

        client = TestClient(requester_class=RequesterCheckArtifactProperties,
                            servers=servers, users={"default": [("lasote", "mypass")]})
        _create_property_files(client, wanted_vars)

        client.save({"conanfile.py": GenConanfile("Hello0", "0.1")})
        client.run("export . lasote/stable")
        client.run("upload Hello0/0.1@lasote/stable -c")

    def test_matrix_params(self):
        test_server = TestServer(server_capabilities=[MATRIX_PARAMS, ])
        servers = {"default": test_server}

        wanted_vars = {"key0": "value",
                       "key1": "with space",
                       "key2": "with/slash",
                       "key3": "with.dot",
                       "key4": "with;semicolon",
                       "key5": "with~virgul",
                       "key6": "with#hash"
                       }

        class RequesterCheckArtifactProperties(TestRequester):
            def put(self_requester, url, **kwargs):
                # Check headers
                self.assertListEqual(list(kwargs["headers"].keys()),
                                     ["X-Checksum-Sha1", "User-Agent"])

                # Check matrix params
                m = re.match(r"^[^;\s]+;(?P<matrix_params>[^/]+)/.*", url)
                mp = m.group("matrix_params")
                values = [it.split("=") for it in mp.split(';')]
                values = {key: unquote(value) for (key, value) in values}
                for name, value in wanted_vars.items():
                    value1 = values[name]
                    self.assertEqual(value1, value)

                slice_start, slice_end = m.span(1)
                url = url[: slice_start-1] + url[slice_end:]  # matrix params are not implemented for conan-server
                return super(RequesterCheckArtifactProperties, self_requester).put(url, **kwargs)

        client = TestClient(requester_class=RequesterCheckArtifactProperties,
                            servers=servers, users={"default": [("lasote", "mypass")]})
        _create_property_files(client, wanted_vars)

        client.save({"conanfile.py": GenConanfile("Hello0", "0.1")})
        client.run("export . lasote/stable")
        client.run("upload Hello0/0.1@lasote/stable -c")


def _create_property_files(client, values):
    lines = ["#Some comment", " #Some comments"]
    for name, value in values.items():
        lines.append("%s=%s" % (name, value))
    save(client.cache.artifacts_properties_path, "\n".join(lines))



import json
import os
import unittest

from mock import Mock

from conans.client.generators.json_generator import JsonGenerator
from conans.model.build_info import CppInfo
from conans.model.conan_file import ConanFile
from conans.model.env_info import EnvValues, EnvInfo
from conans.model.ref import ConanFileReference
from conans.model.settings import Settings
from conans.model.user_info import UserInfo, DepsUserInfo


class JsonTest(unittest.TestCase):

    def test_variables_setup(self):
        conanfile = ConanFile(Mock(), None)
        conanfile.initialize(Settings({}), EnvValues())

        # Add some cpp_info for dependencies
        ref = ConanFileReference.loads("MyPkg/0.1@lasote/stables")
        cpp_info = CppInfo(ref.name, "dummy_root_folder1")
        cpp_info.defines = ["MYDEFINE1"]
        cpp_info.cflags.append("-Flag1=23")
        cpp_info.version = "1.3"
        cpp_info.description = "My cool description"
        cpp_info.build_modules.append("cmake/module.cmake")
        conanfile.deps_cpp_info.add(ref.name, cpp_info)

        ref = ConanFileReference.loads("MyPkg2/0.1@lasote/stables")
        cpp_info = CppInfo(ref.name, "dummy_root_folder2")
        cpp_info.defines = ["MYDEFINE2"]
        cpp_info.version = "2.3"
        cpp_info.exelinkflags = ["-exelinkflag"]
        cpp_info.sharedlinkflags = ["-sharedlinkflag"]
        cpp_info.cxxflags = ["-cxxflag"]
        cpp_info.public_deps = ["MyPkg"]
        conanfile.deps_cpp_info.add(ref.name, cpp_info)

        # Add env_info
        env_info = EnvInfo()
        env_info.VAR1 = "env_info-value1"
        env_info.PATH.append("path-extended")
        conanfile.deps_env_info.update(env_info, "env_info_pkg")

        # Add user_info
        user_info = UserInfo()
        user_info.VAR1 = "user_info-value1"
        conanfile.deps_user_info["user_info_pkg"] = user_info

        # Add user_info_build
        conanfile.user_info_build = DepsUserInfo()
        user_info = UserInfo()
        user_info.VAR1 = "user_info_build-value1"
        conanfile.user_info_build["user_info_build_pkg"] = user_info

        generator = JsonGenerator(conanfile)
        json_out = generator.content
        parsed = json.loads(json_out)

        # Check dependencies
        dependencies = parsed["dependencies"]
        self.assertEqual(len(dependencies), 2)
        my_pkg = dependencies[0]
        self.assertEqual(my_pkg["name"], "MyPkg")
        self.assertEqual(my_pkg["description"], "My cool description")
        self.assertEqual(my_pkg["defines"], ["MYDEFINE1"])
        self.assertListEqual(my_pkg["build_modules"]["cmake"], ["cmake/module.cmake"])
        self.assertListEqual(my_pkg["build_modules"]["cmake_multi"], ["cmake/module.cmake"])
        self.assertListEqual(my_pkg["build_modules"]["cmake_find_package"], ["cmake/module.cmake"])
        self.assertListEqual(my_pkg["build_modules"]["cmake_find_package_multi"],
                             ["cmake/module.cmake"])
        self.assertListEqual(my_pkg["build_modules_paths"]["cmake"],
                             [os.path.join("dummy_root_folder1", "cmake/module.cmake")])
        self.assertListEqual(my_pkg["build_modules_paths"]["cmake_multi"],
                             [os.path.join("dummy_root_folder1", "cmake/module.cmake")])
        self.assertListEqual(my_pkg["build_modules_paths"]["cmake_find_package"],
                             [os.path.join("dummy_root_folder1", "cmake/module.cmake")])
        self.assertListEqual(my_pkg["build_modules_paths"]["cmake_find_package_multi"],
                             [os.path.join("dummy_root_folder1", "cmake/module.cmake")])


        # Check env_info
        env_info = parsed["deps_env_info"]
        self.assertListEqual(sorted(env_info.keys()), sorted(["VAR1", "PATH"]))
        self.assertEqual(env_info["VAR1"], "env_info-value1")
        self.assertListEqual(env_info["PATH"], ["path-extended"])

        # Check user_info
        user_info = parsed["deps_user_info"]
        self.assertListEqual(list(user_info.keys()), ["user_info_pkg"])
        self.assertListEqual(list(user_info["user_info_pkg"].keys()), ["VAR1"])
        self.assertEqual(user_info["user_info_pkg"]["VAR1"], "user_info-value1")

        # Check user_info_build
        user_info_build = parsed["user_info_build"]
        self.assertListEqual(list(user_info_build.keys()), ["user_info_build_pkg"])
        self.assertListEqual(list(user_info_build["user_info_build_pkg"].keys()), ["VAR1"])
        self.assertEqual(user_info_build["user_info_build_pkg"]["VAR1"], "user_info_build-value1")



# coding=utf-8
import textwrap
import unittest

from conans.model.ref import ConanFileReference
from conans.test.assets.genconanfile import GenConanfile
from conans.test.utils.tools import TestClient


class LinkedPackageAsProject(unittest.TestCase):

    def setUp(self):
        self.ref = ConanFileReference.loads('lib/version@user/name')

        self.t = TestClient()
        self.t.save({'conanfile.py': GenConanfile()})
        self.t.run('create . parent/version@user/name')
        conan_package_layout = textwrap.dedent("""\
            [includedirs]
            src/include
            """)
        self.t.save({'conanfile.py': GenConanfile().with_require("parent/version@user/name"),
                     "mylayout": conan_package_layout})
        self.t.run('editable add . {}'.format(self.ref))
        self.assertTrue(self.t.cache.installed_as_editable(self.ref))

    def tearDown(self):
        self.t.run('editable remove {}'.format(self.ref))
        self.assertFalse(self.t.cache.installed_as_editable(self.ref))


class InfoCommandOnLocalWorkspaceTest(LinkedPackageAsProject):
    """ Check that commands info/inspect running over an editable package work"""

    def test_no_args(self):
        self.t.run('info .')
        self.assertIn("conanfile.py\n"
                      "    ID: e94ed0d45e4166d2f946107eaa208d550bf3691e\n"
                      "    BuildID: None\n"
                      "    Context: host\n"
                      "    Requires:\n"
                      "        parent/version@user/name\n", self.t.out)

    def test_only_none(self):
        self.t.run('info . --only None')
        self.assertIn("parent/version@user/name\n"
                      "conanfile.py", self.t.out)

    def test_paths(self):
        self.t.run('info . --paths')
        self.assertIn("conanfile.py\n"
                      "    ID: e94ed0d45e4166d2f946107eaa208d550bf3691e\n"
                      "    BuildID: None\n"
                      "    Context: host\n"
                      "    Requires:\n"
                      "        parent/version@user/name\n", self.t.out)


class InfoCommandUsingReferenceTest(LinkedPackageAsProject):

    def test_no_args(self):
        self.t.run('info {}'.format(self.ref))
        rev = "    Revision: None\n"\
              "    Package revision: None\n" \
              if self.t.cache.config.revisions_enabled else ""  # Project revision is None
        expected = "lib/version@user/name\n" \
                   "    ID: e94ed0d45e4166d2f946107eaa208d550bf3691e\n" \
                   "    BuildID: None\n" \
                   "    Context: host\n" \
                   "    Remote: None\n" \
                   "    Provides: lib\n" \
                   "    Recipe: Editable\n{}" \
                   "    Binary: Editable\n" \
                   "    Binary remote: None\n" \
                   "    Requires:\n" \
                   "        parent/version@user/name\n".format(rev)
        self.assertIn(expected, self.t.out)

    def test_only_none(self):
        self.t.run('info {} --only None'.format(self.ref))
        self.assertListEqual(sorted(str(self.t.out).splitlines()),
                             sorted(["lib/version@user/name", "parent/version@user/name"]))


def test_info_paths():
    # https://github.com/conan-io/conan/issues/7054
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conans import ConanFile
        class Pkg(ConanFile):
            def layout(self):
                self.folders.source = "."
                self.folders.build = "."
        """)
    c.save({"pkg/conanfile.py": conanfile,
            "consumer/conanfile.py": GenConanfile().with_require("pkg/0.1")})
    c.run("editable add pkg pkg/0.1@")
    c.run("info consumer --paths")
    # TODO: Conan 2.0: see if it is possible to get the full actual values
    assert "export_folder:" in c.out  # Important bit is it doesn't raise an error



import os
import tempfile
import textwrap
import unittest

import pytest
import yaml
from mock import Mock

from conans.client.graph.python_requires import ConanPythonRequire
from conans.client.loader import ConanFileLoader
from conans.client.tools.env import environment_append
from conans.model.ref import ConanFileReference
from conans.paths import DATA_YML
from conans.test.utils.tools import TestClient
from conans.util.files import load
from conans.util.files import save_files


class SCMDataToConanDataTestCase(unittest.TestCase):
    ref = ConanFileReference.loads("name/version@")

    def test_plain_recipe(self):
        conanfile = textwrap.dedent("""
            import os
            from conans import ConanFile

            class Recipe(ConanFile):
                scm = {"type": "git", "url": "myurl", "revision": "myrev",
                       "username": "myuser", "password": os.environ.get("SECRET", None),
                       "verify_ssl": False, "shallow": False}
        """)
        t = TestClient()
        t.save({'conanfile.py': conanfile})
        t.run("config set general.scm_to_conandata=1")
        t.run("export . name/version@")

        # Check exported files
        package_layout = t.cache.package_layout(self.ref)
        exported_conanfile = load(package_layout.conanfile())
        self.assertEqual(exported_conanfile, conanfile)
        exported_conandata = load(os.path.join(package_layout.export(), DATA_YML))
        conan_data = yaml.safe_load(exported_conandata)
        self.assertDictEqual(conan_data['.conan']['scm'],
                             {"type": "git", "url": "myurl", "revision": "myrev",
                              "verify_ssl": False, "shallow": False})

        # Check the recipe gets the proper values
        t.run("inspect name/version@ -a scm")
        self.assertIn("password: None", t.out)
        self.assertIn("username: myuser", t.out)
        with environment_append({"SECRET": "42"}):
            t.run("inspect name/version@ -a scm")
            self.assertIn("password: 42", t.out)
            self.assertIn("username: myuser", t.out)

    def test_save_special_chars(self):
        conanfile = textwrap.dedent("""
            import os
            from conans import ConanFile

            class Recipe(ConanFile):
                scm = {"type": "git", "url": 'weir"d', "revision": 123,
                       "subfolder": "weir\\"d", "submodule": "don't"}
        """)
        t = TestClient()
        t.save({'conanfile.py': conanfile})
        t.run("config set general.scm_to_conandata=1")
        t.run("export . name/version@")

        # Check exported files
        package_layout = t.cache.package_layout(self.ref)
        exported_conanfile = load(package_layout.conanfile())
        self.assertEqual(exported_conanfile, conanfile)
        exported_conandata = load(os.path.join(package_layout.export(), DATA_YML))
        conan_data = yaml.safe_load(exported_conandata)
        self.assertDictEqual(conan_data['.conan']['scm'], {"type": "git", "url": 'weir"d',
                                                           "revision": 123, "subfolder": "weir\"d",
                                                           "submodule": "don't"})

        # Check the recipe gets the proper values
        t.run("inspect name/version@ -a scm")
        self.assertIn("revision: 123", t.out)
        self.assertIn('subfolder: weir"d', t.out)
        self.assertIn("submodule: don't", t.out)
        self.assertIn("revision: 123", t.out)
        self.assertIn('url: weir"d', t.out)

    @pytest.mark.tool_git
    def test_auto_is_replaced(self):
        conanfile = textwrap.dedent("""
            import os
            from conans import ConanFile

            class Recipe(ConanFile):
                scm = {"type": "git", "url": "auto", "revision": "auto"}
        """)
        t = TestClient()
        commit = t.init_git_repo({'conanfile.py': conanfile})
        t.run_command('git remote add origin https://myrepo.com.git')
        t.run("config set general.scm_to_conandata=1")
        t.run("export . name/version@")

        # Check exported files
        package_layout = t.cache.package_layout(self.ref)
        exported_conanfile = load(package_layout.conanfile())
        self.assertEqual(exported_conanfile, conanfile)
        exported_conandata = load(os.path.join(package_layout.export(), DATA_YML))
        conan_data = yaml.safe_load(exported_conandata)
        self.assertDictEqual(conan_data['.conan']['scm'],
                             {"type": "git", "url": 'https://myrepo.com.git', "revision": commit})

        # Check the recipe gets the proper values
        t.run("inspect name/version@ -a scm")
        self.assertIn("revision: {}".format(commit), t.out)
        self.assertIn("type: git", t.out)
        self.assertIn("url: https://myrepo.com.git", t.out)

    def test_existing_field(self):
        conanfile = textwrap.dedent("""
            import os
            from conans import ConanFile

            class Recipe(ConanFile):
                scm = {"type": "git", "url": "myurl", "revision": "myrev",
                       "username": "myuser", "password": os.environ.get("SECRET", None),}
        """)
        t = TestClient()
        t.save({'conanfile.py': conanfile,
                DATA_YML: yaml.safe_dump({'.conan': {'scm_data': {}}}, default_flow_style=False)})

        # Without activating the behavior, it works
        t.run("export . name/version@")

        # It fails with it activated
        t.run("config set general.scm_to_conandata=1")
        t.run("export . name/version@", assert_error=True)
        self.assertIn("ERROR: Field '.conan' inside 'conandata.yml' file is"
                      " reserved to Conan usage.", t.out)

    @pytest.mark.tool_git
    def test_empty_conandata(self):
        # https://github.com/conan-io/conan/issues/8209
        conanfile = textwrap.dedent("""
            from conans import ConanFile

            class Recipe(ConanFile):
                scm = {"type": "git", "url": "auto", "revision": "auto"}
            """)
        t = TestClient()
        commit = t.init_git_repo({'conanfile.py': conanfile,
                                  'conandata.yml': ""})
        t.run_command('git remote add origin https://myrepo.com.git')
        t.run("config set general.scm_to_conandata=1")
        t.run("export . name/version@")

        # Check exported files
        package_layout = t.cache.package_layout(self.ref)
        exported_conanfile = load(package_layout.conanfile())
        self.assertEqual(exported_conanfile, conanfile)
        exported_conandata = load(os.path.join(package_layout.export(), DATA_YML))
        conan_data = yaml.safe_load(exported_conandata)
        self.assertDictEqual(conan_data['.conan']['scm'],
                             {"type": "git", "url": 'https://myrepo.com.git', "revision": commit})


class ParseSCMFromConanDataTestCase(unittest.TestCase):
    loader = ConanFileLoader(runner=None, output=Mock(),
                             python_requires=ConanPythonRequire(None, None))

    def test_parse_data(self):
        conanfile = textwrap.dedent("""
            import os
            from conans import ConanFile

            class Recipe(ConanFile):
                scm = {"type": "git", "url": "auto", "revision": "auto"}
        """)
        conan_data = {
            'something_else': {},
            '.conan': {'scm': {
                'type': 'git',
                'url': 'http://myrepo.com',
                'shallow': False,
                'revision': 'abcdefghijk'
            }}}
        test_folder = tempfile.mkdtemp()
        save_files(test_folder, {'conanfile.py': conanfile,
                                 DATA_YML: yaml.safe_dump(conan_data, default_flow_style=False)})

        conanfile_path = os.path.join(test_folder, 'conanfile.py')
        conanfile, _ = self.loader.load_basic_module(conanfile_path=conanfile_path)
        self.assertDictEqual(conanfile.scm, conan_data['.conan']['scm'])


@pytest.mark.tool_git
def test_auto_can_be_automated():
    # https://github.com/conan-io/conan/issues/8881
    conanfile = textwrap.dedent("""
        import os
        from conans import ConanFile

        class Recipe(ConanFile):
            scm = {"type": "git",
                   "url": os.getenv("USER_EXTERNAL_URL", "auto"),
                   "revision": os.getenv("USER_EXTERNAL_COMMIT", "auto")}
    """)
    t = TestClient()
    commit = t.init_git_repo({'conanfile.py': conanfile,
                              'conandata.yml': ""})
    t.run_command('git remote add origin https://myrepo.com.git')
    t.run("config set general.scm_to_conandata=1")
    t.run("export . pkg/1.0@")

    def _check(client):
        # Check exported files
        ref = ConanFileReference.loads("pkg/1.0")
        package_layout = client.cache.package_layout(ref)
        exported_conanfile = load(package_layout.conanfile())
        assert exported_conanfile == conanfile
        exported_conandata = load(os.path.join(package_layout.export(), DATA_YML))
        conan_data = yaml.safe_load(exported_conandata)
        assert conan_data['.conan']['scm'] == {"type": "git",
                                               "url": 'https://myrepo.com.git',
                                               "revision": commit}

    _check(t)

    # Now try from another folder, doing a copy and using the env-vars
    t = TestClient(default_server_user=True)
    t.run("config set general.scm_to_conandata=1")
    t.save({"conanfile.py": conanfile})
    with environment_append({"USER_EXTERNAL_URL": "https://myrepo.com.git",
                             "USER_EXTERNAL_COMMIT": commit}):
        t.run("export . pkg/1.0@")
    _check(t)

    t.run("upload * -c")
    t.run("remove * -f")
    t.run("install pkg/1.0@ --build", assert_error=True)
    assert "pkg/1.0: SCM: Getting sources from url: 'https://myrepo.com.git'" in t.out

    with environment_append({"USER_EXTERNAL_URL": "https://other.different.url",
                             "USER_EXTERNAL_COMMIT": "invalid commit"}):
        t.run("install pkg/1.0@ --build", assert_error=True)
        print(t.out)
        assert "pkg/1.0: SCM: Getting sources from url: 'https://myrepo.com.git'" in t.out



import platform
import textwrap

import pytest

from conans.model.ref import ConanFileReference
from conans.test.assets.genconanfile import GenConanfile
from conans.test.assets.sources import gen_function_h, gen_function_cpp
from conans.test.utils.tools import TestClient


@pytest.mark.slow
@pytest.mark.tool_cmake
@pytest.fixture(scope="module")
def setup_client_with_greetings():
    """
    creates a multi-component package with 2 components "hello" and "bye
    """
    hello_h = gen_function_h(name="hello")
    hello_cpp = gen_function_cpp(name="hello", includes=["hello"])
    bye_h = gen_function_h(name="bye")
    bye_cpp = gen_function_cpp(name="bye", includes=["bye"])

    conanfile_greetings = textwrap.dedent("""
        from conans import ConanFile
        from conan.tools.cmake import CMake

        class GreetingsConan(ConanFile):
            name = "greetings"
            version = "0.0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeDeps", "CMakeToolchain"
            exports_sources = "src/*"
            options = {"components": ["standard", "custom", "none"]}
            default_options = {"components": "standard"}

            def build(self):
                cmake = CMake(self)
                cmake.configure(build_script_folder="src")
                cmake.build()

            def package(self):
                self.copy("*.h", dst="include", src="src")
                self.copy("*.lib", dst="lib", keep_path=False)
                self.copy("*.a", dst="lib", keep_path=False)

            def package_info(self):
                if self.options.components == "standard":
                    self.cpp_info.components["hello"].libs = ["hello"]
                    self.cpp_info.components["bye"].libs = ["bye"]
                elif self.options.components == "custom":
                    self.cpp_info.set_property("cmake_file_name", "MYG")
                    self.cpp_info.set_property("cmake_target_name", "MyGreetings::MyGreetings")

                    self.cpp_info.components["hello"].set_property("cmake_target_name", "MyGreetings::MyHello")
                    self.cpp_info.components["bye"].set_property("cmake_target_name", "MyGreetings::MyBye")

                    self.cpp_info.components["hello"].libs = ["hello"]
                    self.cpp_info.components["bye"].libs = ["bye"]
                else:
                    self.cpp_info.libs = ["hello", "bye"]

            def package_id(self):
                del self.info.options.components
        """)

    cmakelists_greetings = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(greetings CXX)

        add_library(hello hello.cpp)
        add_library(bye bye.cpp)
        """)

    test_package_greetings_conanfile = textwrap.dedent("""
        import os
        from conans import ConanFile
        from conan.tools.cmake import CMake

        class GreetingsTestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeDeps", "CMakeToolchain"
            requires = "greetings/0.0.1"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def test(self):
                path = "{}".format(self.settings.build_type) if self.settings.os == "Windows" else "."
                self.run("{}{}example".format(path, os.sep))
        """)
    test_package_greetings_cpp = gen_function_cpp(name="main", includes=["hello", "bye"],
                                                  calls=["hello", "bye"])

    test_package_greetings_cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(PackageTest CXX)

        find_package(greetings)

        add_executable(example example.cpp)
        target_link_libraries(example greetings::greetings)
        """)

    client = TestClient()
    client.save({"conanfile.py": conanfile_greetings,
                 "src/CMakeLists.txt": cmakelists_greetings,
                 "src/hello.h": hello_h,
                 "src/hello.cpp": hello_cpp,
                 "src/bye.h": bye_h,
                 "src/bye.cpp": bye_cpp,
                 "test_package/conanfile.py": test_package_greetings_conanfile,
                 "test_package/example.cpp": test_package_greetings_cpp,
                 "test_package/CMakeLists.txt": test_package_greetings_cmakelists})
    client.run("create . -s build_type=Release")
    assert "hello: Release!" in client.out
    assert "bye: Release!" in client.out
    client.run("create . -s build_type=Debug")
    assert "hello: Debug!" in client.out
    assert "bye: Debug!" in client.out
    return client


def create_chat(client, components, package_info, cmake_find, test_cmake_find):
    conanfile = textwrap.dedent("""
        from conans import ConanFile
        from conan.tools.cmake import CMake

        class Chat(ConanFile):
            name = "chat"
            version = "0.0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeDeps", "CMakeToolchain"
            exports_sources = "src/*"
            requires = "greetings/0.0.1"
            default_options = {{"greetings:components": "{}"}}

            def build(self):
                cmake = CMake(self)
                cmake.configure(build_script_folder="src")
                cmake.build()

            def package(self):
                self.copy("*.h", dst="include", src="src")
                self.copy("*.lib", dst="lib", keep_path=False)
                self.copy("*.a", dst="lib", keep_path=False)

            def package_info(self):
                {}
        """).format(components, "\n        ".join(package_info.splitlines()))
    sayhello_h = gen_function_h(name="sayhello")
    sayhello_cpp = gen_function_cpp(name="sayhello", includes=["hello"], calls=["hello"])
    sayhellobye_h = gen_function_h(name="sayhellobye")
    sayhellobye_cpp = gen_function_cpp(name="sayhellobye", includes=["sayhello", "bye"],
                                       calls=["sayhello", "bye"])

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(world CXX)

        %s
        """ % cmake_find)

    test_conanfile = textwrap.dedent("""
        import os
        from conans import ConanFile
        from conan.tools.cmake import CMake

        class WorldTestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeDeps", "CMakeToolchain"
            requires = "chat/0.0.1"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def test(self):
                path = "{}".format(self.settings.build_type) if self.settings.os == "Windows" else "."
                self.run("{}{}example".format(path, os.sep))
                self.run("{}{}example2".format(path, os.sep))
        """)
    test_example_cpp = gen_function_cpp(name="main", includes=["sayhellobye"], calls=["sayhellobye"])

    test_cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(PackageTest CXX)

        %s
        """ % test_cmake_find)
    client.save({"conanfile.py": conanfile,
                 "src/CMakeLists.txt": cmakelists,
                 "src/sayhello.h": sayhello_h,
                 "src/sayhello.cpp": sayhello_cpp,
                 "src/sayhellobye.h": sayhellobye_h,
                 "src/sayhellobye.cpp": sayhellobye_cpp,
                 "test_package/conanfile.py": test_conanfile,
                 "test_package/CMakeLists.txt": test_cmakelists,
                 "test_package/example.cpp": test_example_cpp}, clean_first=True)
    client.run("create . -s build_type=Release")
    assert "sayhellobye: Release!" in client.out
    assert "sayhello: Release!" in client.out
    assert "hello: Release!" in client.out
    assert "bye: Release!" in client.out
    client.run("create . -s build_type=Debug")
    assert "sayhellobye: Debug!" in client.out
    assert "sayhello: Debug!" in client.out
    assert "hello: Debug!" in client.out
    assert "bye: Debug!" in client.out


def test_standard_names(setup_client_with_greetings):
    client = setup_client_with_greetings

    package_info = textwrap.dedent("""
        self.cpp_info.components["sayhello"].requires = ["greetings::hello"]
        self.cpp_info.components["sayhello"].libs = ["sayhello"]
        self.cpp_info.components["sayhellobye"].requires = ["sayhello", "greetings::bye"]
        self.cpp_info.components["sayhellobye"].libs = ["sayhellobye"]
        """)
    cmake_find = textwrap.dedent("""
        find_package(greetings COMPONENTS hello bye)

        add_library(sayhello sayhello.cpp)
        target_link_libraries(sayhello greetings::hello)

        add_library(sayhellobye sayhellobye.cpp)
        target_link_libraries(sayhellobye sayhello greetings::bye)
        """)
    test_cmake_find = textwrap.dedent("""
        find_package(chat)

        add_executable(example example.cpp)
        target_link_libraries(example chat::sayhellobye)

        add_executable(example2 example.cpp)
        target_link_libraries(example2 chat::chat)
        """)

    create_chat(client, "standard", package_info, cmake_find, test_cmake_find)

    # Test consumer multi-config
    if platform.system() == "Windows":
        with client.chdir("test_package"):
            client.run("install . -s build_type=Release")
            client.run("install . -s build_type=Debug")
            client.run_command('cmake . -DCMAKE_TOOLCHAIN_FILE=conan_toolchain.cmake')
            client.run_command("cmake --build . --config Debug")
            client.run_command(r".\Debug\example.exe")
            assert "sayhellobye: Debug!" in client.out
            assert "sayhello: Debug!" in client.out
            assert "hello: Debug!" in client.out
            assert "bye: Debug!" in client.out
            client.run_command("cmake --build . --config Release")
            client.run_command(r".\Release\example.exe")
            assert "sayhellobye: Release!" in client.out
            assert "sayhello: Release!" in client.out
            assert "hello: Release!" in client.out
            assert "bye: Release!" in client.out


def test_custom_names(setup_client_with_greetings):
    client = setup_client_with_greetings

    package_info = textwrap.dedent("""
        # NOTE: For the new CMakeDeps only filenames mean filename, it is not using the "names" field
        self.cpp_info.set_property("cmake_target_name", "MyChat::MyChat")
        self.cpp_info.set_property("cmake_file_name", "MyChat")

        self.cpp_info.components["sayhello"].set_property("cmake_target_name", "MyChat::MySay")

        self.cpp_info.components["sayhello"].requires = ["greetings::hello"]
        self.cpp_info.components["sayhello"].libs = ["sayhello"]
        self.cpp_info.components["sayhellobye"].set_property("cmake_target_name", "MyChat::MySayBye")

        self.cpp_info.components["sayhellobye"].requires = ["sayhello", "greetings::bye"]
        self.cpp_info.components["sayhellobye"].libs = ["sayhellobye"]
        """)

    cmake_find = textwrap.dedent("""
        find_package(MYG COMPONENTS MyHello MyBye)

        add_library(sayhello sayhello.cpp)
        target_link_libraries(sayhello MyGreetings::MyHello)

        add_library(sayhellobye sayhellobye.cpp)
        target_link_libraries(sayhellobye sayhello MyGreetings::MyBye)
        """)

    test_cmake_find = textwrap.dedent("""
        find_package(MyChat)

        add_executable(example example.cpp)
        target_link_libraries(example MyChat::MySayBye)

        add_executable(example2 example.cpp)
        target_link_libraries(example2 MyChat::MyChat)
        """)
    create_chat(client, "custom", package_info, cmake_find, test_cmake_find)


def test_different_namespace(setup_client_with_greetings):
    client = setup_client_with_greetings

    package_info = textwrap.dedent("""
        self.cpp_info.set_property("cmake_target_name", "MyChat::MyGlobalChat")
        self.cpp_info.set_property("cmake_file_name", "MyChat")

        self.cpp_info.components["sayhello"].set_property("cmake_target_name", "MyChat::MySay")

        self.cpp_info.components["sayhello"].requires = ["greetings::hello"]
        self.cpp_info.components["sayhello"].libs = ["sayhello"]
        self.cpp_info.components["sayhellobye"].set_property("cmake_target_name", "MyChat::MySayBye")

        self.cpp_info.components["sayhellobye"].requires = ["sayhello", "greetings::bye"]
        self.cpp_info.components["sayhellobye"].libs = ["sayhellobye"]
        """)

    cmake_find = textwrap.dedent("""
        find_package(MYG COMPONENTS MyHello MyBye)

        add_library(sayhello sayhello.cpp)
        target_link_libraries(sayhello MyGreetings::MyHello)

        add_library(sayhellobye sayhellobye.cpp)
        target_link_libraries(sayhellobye sayhello MyGreetings::MyBye)
        """)

    test_cmake_find = textwrap.dedent("""
        find_package(MyChat)

        add_executable(example example.cpp)
        target_link_libraries(example MyChat::MySayBye)

        add_executable(example2 example.cpp)
        target_link_libraries(example2 MyChat::MyGlobalChat)
        """)
    create_chat(client, "custom", package_info, cmake_find, test_cmake_find)



def test_no_components(setup_client_with_greetings):
    client = setup_client_with_greetings

    package_info = textwrap.dedent("""
        self.cpp_info.components["sayhello"].requires = ["greetings::greetings"]
        self.cpp_info.components["sayhello"].libs = ["sayhello"]
        self.cpp_info.components["sayhellobye"].requires = ["sayhello", "greetings::greetings"]
        self.cpp_info.components["sayhellobye"].libs = ["sayhellobye"]
        """)

    cmake_find = textwrap.dedent("""
        find_package(greetings)

        add_library(sayhello sayhello.cpp)
        target_link_libraries(sayhello greetings::greetings)

        add_library(sayhellobye sayhellobye.cpp)
        target_link_libraries(sayhellobye sayhello greetings::greetings)
        """)

    test_cmake_find = textwrap.dedent("""
        find_package(chat)

        add_executable(example example.cpp)
        target_link_libraries(example chat::sayhellobye)

        add_executable(example2 example.cpp)
        target_link_libraries(example2 chat::chat)
        """)
    create_chat(client, "none", package_info, cmake_find, test_cmake_find)


@pytest.mark.slow
@pytest.mark.tool_cmake
def test_same_names():
    client = TestClient()
    conanfile_greetings = textwrap.dedent("""
        from conans import ConanFile
        from conan.tools.cmake import CMake

        class HelloConan(ConanFile):
            name = "hello"
            version = "0.0.1"
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeDeps", "CMakeToolchain"
            exports_sources = "src/*"

            def build(self):
                cmake = CMake(self)
                cmake.configure(build_script_folder="src")
                cmake.build()

            def package(self):
                self.copy("*.h", dst="include", src="src")
                self.copy("*.lib", dst="lib", keep_path=False)
                self.copy("*.a", dst="lib", keep_path=False)

            def package_info(self):
                self.cpp_info.components["global"].name = "hello"
                self.cpp_info.components["global"].libs = ["hello"]
        """)
    hello_h = gen_function_h(name="hello")
    hello_cpp = gen_function_cpp(name="hello", includes=["hello"])

    cmakelists_greetings = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(greetings CXX)

        add_library(hello hello.cpp)
        """)
    test_package_greetings_conanfile = textwrap.dedent("""
        import os
        from conans import ConanFile
        from conan.tools.cmake import CMake

        class HelloTestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeDeps", "CMakeToolchain"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def test(self):
                path = "{}".format(self.settings.build_type) if self.settings.os == "Windows" else "."
                self.run("{}{}example".format(path, os.sep))
        """)
    test_package_greetings_cpp = gen_function_cpp(name="main", includes=["hello"], calls=["hello"])

    test_package_greetings_cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.0)
        project(PackageTest CXX)

        find_package(hello)

        add_executable(example example.cpp)
        target_link_libraries(example hello::hello)
        """)
    client.save({"conanfile.py": conanfile_greetings,
                 "src/CMakeLists.txt": cmakelists_greetings,
                 "src/hello.h": hello_h,
                 "src/hello.cpp": hello_cpp,
                 "test_package/conanfile.py": test_package_greetings_conanfile,
                 "test_package/example.cpp": test_package_greetings_cpp,
                 "test_package/CMakeLists.txt": test_package_greetings_cmakelists})
    client.run("create .")
    assert "hello: Release!" in client.out


@pytest.mark.tool_cmake
class TestComponentsCMakeGenerators:

    def test_component_not_found(self):
        conanfile = textwrap.dedent("""
            from conans import ConanFile
            class GreetingsConan(ConanFile):
                def package_info(self):
                    self.cpp_info.components["hello"].libs = ["hello"]
        """)
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . greetings/0.0.1@")

        conanfile = textwrap.dedent("""
            from conans import ConanFile
            class WorldConan(ConanFile):
                requires = "greetings/0.0.1"
                def package_info(self):
                    self.cpp_info.components["helloworld"].requires = ["greetings::non-existent"]
        """)
        client.save({"conanfile.py": conanfile})
        client.run("create . world/0.0.1@")
        client.run("install world/0.0.1@ -g CMakeDeps", assert_error=True)
        assert ("Component 'greetings::non-existent' not found in 'greetings' "
                "package requirement" in client.out)

    def test_component_not_found_same_name_as_pkg_require(self):
        zlib = GenConanfile("zlib", "0.1").with_setting("build_type").with_generator("CMakeDeps")
        mypkg = GenConanfile("mypkg", "0.1").with_setting("build_type").with_generator("CMakeDeps")
        final = GenConanfile("final", "0.1").with_setting("build_type").with_generator("CMakeDeps")\
            .with_require(ConanFileReference("zlib", "0.1", None, None))\
            .with_require(ConanFileReference("mypkg", "0.1", None, None))\
            .with_package_info(cpp_info={"components": {"cmp": {"requires": ["mypkg::zlib",
                                                                             "zlib::zlib"]}}},
                               env_info={})
        consumer = GenConanfile("consumer", "0.1").with_setting("build_type")\
            .with_generator("CMakeDeps")\
            .with_requirement(ConanFileReference("final", "0.1", None, None))

        consumer = textwrap.dedent("""
            from conans import ConanFile
            from conan.tools.cmake import CMakeDeps
            class HelloConan(ConanFile):
                name = 'consumer'
                version = '0.1'

                def generate(self):
                    deps = CMakeDeps(self)
                    deps.check_components_exist = True
                    deps.generate()

                def requirements(self):
                    self.requires("final/0.1")

                settings = "build_type"
            """)

        def test_component_not_found(self):
            conanfile = textwrap.dedent("""
                from conans import ConanFile
                class GreetingsConan(ConanFile):
                    def package_info(self):
                        self.cpp_info.components["hello"].libs = ["hello"]
            """)
            client = TestClient()
            client.save({"conanfile.py": conanfile})
            client.run("create . greetings/0.0.1@")

            conanfile = textwrap.dedent("""
                from conans import ConanFile
                class WorldConan(ConanFile):
                    requires = "greetings/0.0.1"
                    def package_info(self):
                        self.cpp_info.components["helloworld"].requires = ["greetings::non-existent"]
            """)
            client.save({"conanfile.py": conanfile})
            client.run("create . world/0.0.1@")
            client.run("install world/0.0.1@ -g CMakeDeps", assert_error=True)
            assert ("Component 'greetings::non-existent' not found in 'greetings' "
                    "package requirement" in client.out)

        client = TestClient()
        client.save({"zlib.py": zlib, "mypkg.py": mypkg, "final.py": final, "consumer.py": consumer})
        client.run("create zlib.py")
        client.run("create mypkg.py")
        client.run("create final.py")
        client.run("install consumer.py", assert_error=True)
        assert "Component 'mypkg::zlib' not found in 'mypkg' package requirement" in client.out

    @pytest.mark.slow
    def test_same_name_global_target_collision(self):
        # https://github.com/conan-io/conan/issues/7889
        conanfile_tpl = textwrap.dedent("""
            from conans import ConanFile
            from conan.tools.cmake import CMake

            class Conan(ConanFile):
                name = "{name}"
                version = "1.0"
                settings = "os", "compiler", "build_type", "arch"
                generators = "CMakeDeps", "CMakeToolchain"
                exports_sources = "src/*"

                def build(self):
                    cmake = CMake(self)
                    cmake.configure(build_script_folder="src")
                    cmake.build()

                def package(self):
                    self.copy("*.h", dst="include", src="src")
                    self.copy("*.lib", dst="lib", keep_path=False)
                    self.copy("*.a", dst="lib", keep_path=False)

                def package_info(self):
                    self.cpp_info.set_property("cmake_target_name", "nonstd::nonstd" )
                    self.cpp_info.set_property("cmake_file_name", "{name}")

                    self.cpp_info.components["1"].set_property("cmake_target_name", "nonstd::{name}")
                    self.cpp_info.components["1"].libs = ["{name}"]
            """)
        basic_cmake = textwrap.dedent("""
            cmake_minimum_required(VERSION 3.15)
            set(CMAKE_CXX_COMPILER_WORKS 1)
            set(CMAKE_CXX_ABI_COMPILED 1)
            project(middle CXX)
            add_library({name} {name}.cpp)
            """)
        client = TestClient()
        for name in ["expected", "variant"]:
            client.run("new {name}/1.0 -s".format(name=name))
            client.save({"conanfile.py": conanfile_tpl.format(name=name),
                         "src/CMakeLists.txt": basic_cmake.format(name=name)})
            client.run("create . {name}/1.0@".format(name=name))
        middle_cmakelists = textwrap.dedent("""
            cmake_minimum_required(VERSION 3.15)
            set(CMAKE_CXX_COMPILER_WORKS 1)
            set(CMAKE_CXX_ABI_COMPILED 1)
            project(middle CXX)

            find_package(expected)
            find_package(variant)

            add_library(middle middle.cpp)
            target_link_libraries(middle nonstd::nonstd)
            """)
        middle_h = gen_function_h(name="middle")
        middle_cpp = gen_function_cpp(name="middle", includes=["middle", "expected", "variant"],
                                      calls=["expected", "variant"])
        middle_conanfile = textwrap.dedent("""
            from conans import ConanFile
            from conan.tools.cmake import CMake

            class Conan(ConanFile):
                name = "middle"
                version = "1.0"
                settings = "os", "compiler", "build_type", "arch"
                generators = "CMakeDeps", "CMakeToolchain"
                exports_sources = "src/*"
                requires = "expected/1.0", "variant/1.0"

                def build(self):
                    cmake = CMake(self)
                    cmake.configure(build_script_folder="src")
                    cmake.build()

                def package(self):
                    self.copy("*.h", dst="include", src="src")
                    self.copy("*.lib", dst="lib", keep_path=False)
                    self.copy("*.a", dst="lib", keep_path=False)

                def package_info(self):
                    self.cpp_info.libs = ["middle"]
            """)
        client.save({"conanfile.py": middle_conanfile, "src/CMakeLists.txt": middle_cmakelists,
                     "src/middle.h": middle_h, "src/middle.cpp": middle_cpp}, clean_first=True)
        client.run("create . middle/1.0@")
        conanfile = textwrap.dedent("""
            import os
            from conans import ConanFile
            from conan.tools.cmake import CMake, cmake_layout

            class Conan(ConanFile):
                name = "consumer"
                version = "1.0"
                generators = "CMakeDeps", "CMakeToolchain"
                settings = "os", "compiler", "build_type", "arch"
                exports_sources = "src/*"
                requires = "middle/1.0"

                def layout(self):
                    cmake_layout(self)

                def build(self):
                    cmake = CMake(self)
                    cmake.configure(build_script_folder="src")
                    cmake.build()
                    cmd = os.path.join(self.cpp.build.bindirs[0], "main")
                    self.run(cmd, env="conanrun")
            """)
        cmakelists = textwrap.dedent("""
            cmake_minimum_required(VERSION 3.15)
            set(CMAKE_CXX_COMPILER_WORKS 1)
            set(CMAKE_CXX_ABI_COMPILED 1)
            project(consumer CXX)

            find_package(middle)

            get_target_property(tmp middle::middle INTERFACE_LINK_LIBRARIES)
            message("Middle link libraries: ${tmp}")

            add_executable(main main.cpp)
            target_link_libraries(main middle::middle)
            """)
        main_cpp = gen_function_cpp(name="main", includes=["middle"], calls=["middle"])
        client.save({"conanfile.py": conanfile,
                     "src/CMakeLists.txt": cmakelists,
                     "src/main.cpp": main_cpp}, clean_first=True)
        client.run("create . consumer/1.0@")

        assert 'main: Release!' in client.out
        assert 'middle: Release!' in client.out
        assert 'expected/1.0: Hello World Release!' in client.out
        assert 'variant/1.0: Hello World Release!' in client.out


@pytest.mark.parametrize("check_components_exist", [False, True, None])
def test_targets_declared_in_build_modules(check_components_exist):
    """If a require is declaring the component targets in a build_module, CMakeDeps is
       fine with it, not needed to locate it as a conan declared component"""

    client = TestClient()
    conanfile_hello = str(GenConanfile().with_name("hello").with_version("1.0")
                          .with_exports_sources("*.cmake", "*.h"))
    conanfile_hello += """
    def package(self):
        self.copy("*.h", dst="include")
        self.copy("*.cmake", dst="cmake")

    def package_info(self):
        self.cpp_info.set_property("cmake_build_modules", ["cmake/my_modules.cmake"])
    """
    my_modules = textwrap.dedent("""
    add_library(cool_component INTERFACE)
    target_include_directories(cool_component INTERFACE ${CMAKE_CURRENT_LIST_DIR}/../include/)
    add_library(hello::invented ALIAS cool_component)
    """)
    hello_h = "int cool_header_only=1;"
    client.save({"conanfile.py": conanfile_hello,
                 "my_modules.cmake": my_modules, "hello.h": hello_h})
    client.run("create .")

    conanfile = textwrap.dedent("""
            from conans import ConanFile
            from conan.tools.cmake import CMake, CMakeDeps

            class HelloConan(ConanFile):
                name = 'app'
                version = '1.0'
                exports_sources = "*.txt", "*.cpp"
                generators = "CMakeToolchain"
                requires = ("hello/1.0", )
                settings = "os", "compiler", "arch", "build_type"

                def generate(self):
                    deps = CMakeDeps(self)
                    {}
                    deps.generate()

                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
                    cmake.build()

    """)
    if check_components_exist is False:
        conanfile = conanfile.format("deps.check_components_exist=False")
    elif check_components_exist is True:
        conanfile = conanfile.format("deps.check_components_exist=True")
    else:
        conanfile = conanfile.format("")

    main_cpp = textwrap.dedent("""
        #include <iostream>
        #include "hello.h"

        int main(){
            std::cout << "cool header value: " << cool_header_only;
            return 0;
        }
        """)

    cmakelist = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        set(CMAKE_C_COMPILER_WORKS 1)
        set(CMAKE_C_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(project CXX)

        find_package(hello COMPONENTS hello::invented missing)
        add_executable(myapp main.cpp)
        target_link_libraries(myapp hello::invented)
    """)
    client.save({"conanfile.py": conanfile,
                 "CMakeLists.txt": cmakelist, "main.cpp": main_cpp})
    client.run("create .", assert_error=check_components_exist)
    assert bool(check_components_exist) == ("Conan: Component 'missing' NOT found in package "
                                            "'hello'" in client.out)

    assert "Conan: Including build module" in client.out
    assert "my_modules.cmake" in client.out
    assert bool(check_components_exist) == ("Conan: Component 'hello::invented' found in package 'hello'"
                                            in client.out)


@pytest.mark.tool_cmake
def test_cmakedeps_targets_no_namespace():
    """
    This test is checking that when we add targets with no namespace for the root cpp_info
    and the components, the targets are correctly generated. Before Conan 1.43, Conan
    only generated targets with namespace
    """
    client = TestClient()
    my_pkg = textwrap.dedent("""
        from conans import ConanFile
        class MyPkg(ConanFile):
            name = "my_pkg"
            version = "0.1"
            settings = "os", "arch", "compiler", "build_type"
            def package_info(self):
                self.cpp_info.set_property("cmake_target_name", "nonamespacepkg")
                self.cpp_info.components["MYPKGCOMP"].set_property("cmake_target_name", "MYPKGCOMPNAME")
        """)
    client.save({"my_pkg/conanfile.py": my_pkg}, clean_first=True)
    client.run("create my_pkg")

    conanfile = textwrap.dedent("""
        from conans import ConanFile
        class LibcurlConan(ConanFile):
            name = "libcurl"
            version = "0.1"
            requires = "my_pkg/0.1"
            settings = "os", "arch", "compiler", "build_type"
            def package_info(self):
                self.cpp_info.set_property("cmake_target_name", "CURL")
                self.cpp_info.set_property("cmake_file_name", "CURLFILENAME")
                self.cpp_info.components["curl"].set_property("cmake_target_name", "libcurl")
                self.cpp_info.components["curl2"].set_property("cmake_target_name", "libcurl2")
                self.cpp_info.components["curl2"].requires.extend(["curl", "my_pkg::MYPKGCOMP"])
        """)
    client.save({"libcurl/conanfile.py": conanfile})
    client.run("create libcurl")

    conanfile = textwrap.dedent("""
        from conans import ConanFile
        from conan.tools.cmake import CMakeDeps, CMake, CMakeToolchain
        class Consumer(ConanFile):
            name = "consumer"
            version = "0.1"
            requires = "libcurl/0.1"
            settings = "os", "arch", "compiler", "build_type"
            exports_sources = "CMakeLists.txt"
            def generate(self):
                deps = CMakeDeps(self)
                deps.check_components_exist=True
                deps.generate()
                tc = CMakeToolchain(self)
                tc.generate()
            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
        """)

    cmakelists = textwrap.dedent("""cmake_minimum_required(VERSION 3.15)
        project(Consumer)
        find_package(CURLFILENAME CONFIG REQUIRED COMPONENTS libcurl libcurl2)
        """)

    client.save({"consumer/conanfile.py": conanfile, "consumer/CMakeLists.txt": cmakelists})
    client.run("create consumer")
    assert "Component target declared 'libcurl'" in client.out
    assert "Component target declared 'libcurl2'" in client.out
    assert "Target declared 'CURL'" in client.out
    assert "Component target declared 'MYPKGCOMPNAME'" in client.out
    assert "Target declared 'nonamespacepkg'" in client.out



import unittest
import textwrap
from conans.test.utils.tools import TestClient, GenConanfile


class BuildRequiresFromProfileTest(unittest.TestCase):
    profile_host = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
        compiler=Visual Studio
        compiler.version=16

        [build_requires]
        br2/version
    """)

    profile_build = textwrap.dedent("""
        [settings]
        os=Macos
        arch=x86_64
        compiler=apple-clang
        compiler.version=11.0
        compiler.libcxx=libc++
        build_type=Release

        [build_requires]
        br3/version
    """)

    library_conanfile = textwrap.dedent("""
        from conans import ConanFile

        class Recipe(ConanFile):
            name = "library"
            version = "version"

            build_requires = "br1/version"
    """)

    def test_br_from_profile_host_and_profile_build(self):
        t = TestClient()
        t.save({'profile_host': self.profile_host,
                'profile_build': self.profile_build,
                'library.py': self.library_conanfile,
                'br1.py': GenConanfile(),
                'br2.py': GenConanfile(),
                'br3.py': GenConanfile()})
        t.run("export br1.py br1/version@")
        t.run("export br2.py br2/version@")
        t.run("export br3.py br3/version@")
        t.run("create library.py --profile:host=profile_host --profile:build=profile_build --build")
        self.assertNotIn("br1/version: Applying build-requirement: br2/version", t.out)
        self.assertIn("br1/version: Applying build-requirement: br3/version", t.out)
        self.assertIn("library/version: Applying build-requirement: br2/version", t.out)
        self.assertIn("library/version: Applying build-requirement: br1/version", t.out)


class BuildRequiresContextHostFromProfileTest(unittest.TestCase):
    toolchain = textwrap.dedent("""
        from conans import ConanFile

        class Recipe(ConanFile):
            name = "mytoolchain"
            version = "1.0"
            settings = "os"

            def package_info(self):
                self.env_info.MYTOOLCHAIN_VAR = "MYTOOLCHAIN_VALUE-" + str(self.settings.os)
        """)
    gtest = textwrap.dedent("""
        from conans import ConanFile
        import os

        class Recipe(ConanFile):
            name = "gtest"
            version = "1.0"
            settings = "os"

            def build(self):
                self.output.info("Building with: %s" % os.getenv("MYTOOLCHAIN_VAR"))
                self.output.info("Build OS=%s" % self.settings.os)
            def package_info(self):
                self.output.info("PackageInfo OS=%s" % self.settings.os)
        """)
    library_conanfile = textwrap.dedent("""
         from conans import ConanFile
         import os

         class Recipe(ConanFile):
             name = "library"
             version = "version"
             settings = "os"

             def build_requirements(self):
                self.build_requires("gtest/1.0", force_host_context=True)

             def build(self):
                self.output.info("Building with: %s" % os.getenv("MYTOOLCHAIN_VAR"))
                self.output.info("Build OS=%s" % self.settings.os)
         """)
    profile_host = textwrap.dedent("""
        [settings]
        os = Linux
        [build_requires]
        mytoolchain/1.0
        """)
    profile_build = textwrap.dedent("""
        [settings]
        os = Windows
        """)

    def test_br_from_profile_host_and_profile_build(self):
        t = TestClient()
        t.save({'profile_host': self.profile_host,
                'profile_build': self.profile_build,
                'library.py': self.library_conanfile,
                'mytoolchain.py': self.toolchain,
                "gtest.py": self.gtest})
        t.run("create mytoolchain.py --profile=profile_build")
        t.run("create mytoolchain.py --profile=profile_host")

        # old way, the toolchain will have the same profile (profile_host=Linux) only
        t.run("create gtest.py --profile=profile_host")
        self.assertIn("gtest/1.0: Building with: MYTOOLCHAIN_VALUE-Linux", t.out)
        self.assertIn("gtest/1.0: Build OS=Linux", t.out)

        # new way, the toolchain can now run in Windows, but gtest in Linux
        t.run("create gtest.py --profile=profile_host --profile:build=profile_build")
        self.assertIn("gtest/1.0: Building with: MYTOOLCHAIN_VALUE-Windows", t.out)
        self.assertIn("gtest/1.0: PackageInfo OS=Linux", t.out)

        t.run("create gtest.py --profile=profile_host --profile:build=profile_build --build")
        self.assertIn("gtest/1.0: Building with: MYTOOLCHAIN_VALUE-Windows", t.out)
        self.assertIn("gtest/1.0: Build OS=Linux", t.out)
        self.assertIn("gtest/1.0: PackageInfo OS=Linux", t.out)

        t.run("create library.py --profile:host=profile_host --profile:build=profile_build")
        self.assertIn("gtest/1.0: PackageInfo OS=Linux", t.out)
        self.assertIn("library/version: Build OS=Linux", t.out)
        self.assertIn("library/version: Building with: MYTOOLCHAIN_VALUE-Windows", t.out)

        t.run("create library.py --profile:host=profile_host --profile:build=profile_build --build")
        self.assertIn("gtest/1.0: Build OS=Linux", t.out)
        self.assertIn("gtest/1.0: PackageInfo OS=Linux", t.out)
        self.assertIn("library/version: Build OS=Linux", t.out)
        self.assertIn("library/version: Building with: MYTOOLCHAIN_VALUE-Windows", t.out)


class BuildRequiresBothContextsTest(unittest.TestCase):
    toolchain_creator = textwrap.dedent("""
        from conans import ConanFile

        class Recipe(ConanFile):
            name = "creator"
            version = "1.0"
            settings = "os"

            def package_info(self):
                self.env_info.MYCREATOR_VAR = "MYCREATOR_VALUE-" + str(self.settings.os)
        """)
    toolchain = textwrap.dedent("""
        from conans import ConanFile
        import os

        class Recipe(ConanFile):
            name = "mytoolchain"
            version = "1.0"
            settings = "os"
            def build(self):
                self.output.info("Building with: %s" % os.environ["MYCREATOR_VAR"])
                self.output.info("Build OS=%s" % self.settings.os)

            def package_info(self):
                self.env_info.MYTOOLCHAIN_VAR = "MYTOOLCHAIN_VALUE-" + str(self.settings.os)
        """)
    gtest = textwrap.dedent("""
        from conans import ConanFile
        import os

        class Recipe(ConanFile):
            name = "gtest"
            version = "1.0"
            settings = "os"

            def build(self):
                self.output.info("Building with: %s" % os.environ["MYTOOLCHAIN_VAR"])
                self.output.info("Build OS=%s" % self.settings.os)
            def package_info(self):
                self.output.info("PackageInfo OS=%s" % self.settings.os)
        """)
    library_conanfile = textwrap.dedent("""
         from conans import ConanFile
         import os

         class Recipe(ConanFile):
             name = "library"
             version = "version"
             settings = "os"

             def build_requirements(self):
                self.build_requires("gtest/1.0", force_host_context=True)

             def build(self):
                self.output.info("Building with: %s" % os.environ["MYTOOLCHAIN_VAR"])
                self.output.info("Build OS=%s" % self.settings.os)
         """)
    profile_host = textwrap.dedent("""
        [settings]
        os = Linux
        [build_requires]
        mytoolchain/1.0
        """)
    profile_build = textwrap.dedent("""
        [settings]
        os = Windows
        [build_requires]
        creator/1.0
        """)

    def test_build_requires_both_contexts(self):
        t = TestClient()
        t.save({'profile_host': self.profile_host,
                'profile_build': self.profile_build,
                'library.py': self.library_conanfile,
                'creator.py': self.toolchain_creator,
                'mytoolchain.py': self.toolchain,
                "gtest.py": self.gtest})
        t.run("create creator.py --profile=profile_build")
        t.run("create mytoolchain.py --profile=profile_build")
        self.assertIn("mytoolchain/1.0: Building with: MYCREATOR_VALUE-Windows", t.out)
        self.assertIn("mytoolchain/1.0: Build OS=Windows", t.out)

        # new way, the toolchain can now run in Windows, but gtest in Linux
        t.run("create gtest.py --profile=profile_host --profile:build=profile_build")
        self.assertIn("gtest/1.0: Building with: MYTOOLCHAIN_VALUE-Windows", t.out)
        self.assertIn("gtest/1.0: PackageInfo OS=Linux", t.out)

        # FIXME: This isn't right, it should be CREATOR-Windows, but build_requires
        # FIXME: from profiles are not transitive to build_requires in profiles
        t.run("create gtest.py --profile=profile_host --profile:build=profile_build --build",
              assert_error=True)
        self.assertIn("ERROR: mytoolchain/1.0: Error in build() method, line 10", t.out)
        self.assertIn("KeyError: 'MYCREATOR_VAR'", t.out)

        # Declaring the build_requires in the recipe works, it is just the profile that is
        # not transitive
        toolchain = textwrap.dedent("""
            from conans import ConanFile
            import os

            class Recipe(ConanFile):
                name = "mytoolchain"
                version = "1.0"
                settings = "os"
                build_requires = "creator/1.0"
                def build(self):
                    self.output.info("Building with: %s" % os.environ["MYCREATOR_VAR"])
                    self.output.info("Build OS=%s" % self.settings.os)

                def package_info(self):
                    self.env_info.MYTOOLCHAIN_VAR = "MYTOOLCHAIN_VALUE-" + str(self.settings.os)
            """)
        t.save({'mytoolchain.py': toolchain})
        t.run("create mytoolchain.py --profile=profile_build")
        self.assertIn("mytoolchain/1.0: Building with: MYCREATOR_VALUE-Windows", t.out)
        self.assertIn("mytoolchain/1.0: Build OS=Windows", t.out)

        t.run("create gtest.py --profile=profile_host --profile:build=profile_build --build")
        self.assertIn("mytoolchain/1.0: Building with: MYCREATOR_VALUE-Windows", t.out)
        self.assertIn("mytoolchain/1.0: Build OS=Windows", t.out)
        self.assertIn("gtest/1.0: Building with: MYTOOLCHAIN_VALUE-Windows", t.out)
        self.assertIn("gtest/1.0: Build OS=Linux", t.out)

        t.run("create library.py --profile:host=profile_host --profile:build=profile_build --build")
        self.assertIn("mytoolchain/1.0: Building with: MYCREATOR_VALUE-Windows", t.out)
        self.assertIn("mytoolchain/1.0: Build OS=Windows", t.out)
        self.assertIn("gtest/1.0: Build OS=Linux", t.out)
        self.assertIn("gtest/1.0: PackageInfo OS=Linux", t.out)
        self.assertIn("library/version: Build OS=Linux", t.out)
        self.assertIn("library/version: Building with: MYTOOLCHAIN_VALUE-Windows", t.out)



import os
import textwrap
import unittest

from conans.util.files import load
from conans.model.ref import ConanFileReference, PackageReference
from conans.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID


class TestPackagePythonFiles(unittest.TestCase):
    def test_package_python_files(self):
        client = TestClient(default_server_user=True)
        client.run("config set general.keep_python_files=True")
        conanfile = textwrap.dedent("""
            from conans import ConanFile
            class Pkg(ConanFile):
                exports_sources = "*"
                def package(self):
                    self.copy("*")
            """)
        client.save({"conanfile.py": conanfile,
                     "myfile.pyc": "",
                     "myfile.pyo": "",
                     ".DS_Store": ""})
        client.run("create . pkg/0.1@")
        ref = ConanFileReference.loads("pkg/0.1")
        layout = client.cache.package_layout(ref)
        export = layout.export()
        export_sources = layout.export_sources()
        self.assertTrue(os.path.isfile(os.path.join(export_sources, "myfile.pyc")))
        self.assertTrue(os.path.isfile(os.path.join(export_sources, "myfile.pyo")))
        self.assertTrue(os.path.isfile(os.path.join(export_sources, ".DS_Store")))
        manifest = load(os.path.join(export, "conanmanifest.txt"))
        self.assertIn("myfile.pyc", manifest)
        self.assertIn("myfile.pyo", manifest)
        self.assertNotIn(".DS_Store", manifest)
        pkg_folder = layout.package(PackageReference(ref, NO_SETTINGS_PACKAGE_ID))
        self.assertTrue(os.path.isfile(os.path.join(pkg_folder, "myfile.pyc")))
        self.assertTrue(os.path.isfile(os.path.join(pkg_folder, "myfile.pyo")))
        self.assertTrue(os.path.isfile(os.path.join(pkg_folder, ".DS_Store")))
        manifest = load(os.path.join(pkg_folder, "conanmanifest.txt"))
        self.assertIn("myfile.pyc", manifest)
        self.assertIn("myfile.pyo", manifest)
        self.assertNotIn(".DS_Store", manifest)

        client.run("upload * --all -r=default --confirm")
        client.run("remove * -f")
        client.run("download pkg/0.1@")

        self.assertTrue(os.path.isfile(os.path.join(export_sources, "myfile.pyc")))
        self.assertTrue(os.path.isfile(os.path.join(export_sources, "myfile.pyo")))
        self.assertFalse(os.path.isfile(os.path.join(export_sources, ".DS_Store")))
        manifest = load(os.path.join(export, "conanmanifest.txt"))
        self.assertIn("myfile.pyc", manifest)
        self.assertIn("myfile.pyo", manifest)
        self.assertNotIn(".DS_Store", manifest)
        self.assertTrue(os.path.isfile(os.path.join(pkg_folder, "myfile.pyc")))
        self.assertTrue(os.path.isfile(os.path.join(pkg_folder, "myfile.pyo")))
        self.assertFalse(os.path.isfile(os.path.join(pkg_folder, ".DS_Store")))
        manifest = load(os.path.join(pkg_folder, "conanmanifest.txt"))
        self.assertIn("myfile.pyc", manifest)
        self.assertIn("myfile.pyo", manifest)
        self.assertNotIn(".DS_Store", manifest)



import os
import re

from conans.model.ref import ConanFileReference, PackageReference
from conans.test.assets.genconanfile import GenConanfile
from conans.test.utils.tools import TestClient


def test_auto_package_no_components():
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type")
                     .with_import("from conans import tools")
                     .with_import("from conan.tools.files import AutoPackager"))
    conan_file += """

    def source(self):
        tools.save("source_sources/source_stuff.cpp", "")
        tools.save("source_includes/include1.hpp", "")
        tools.save("source_includes/include2.hpp", "")
        tools.save("source_includes2/include3.h", "")
        tools.save("source_libs/slibone.a", "")
        tools.save("source_libs/slibtwo.a", "")
        tools.save("source_libs/bin_to_discard.exe", "")
        tools.save("source_bins/source_bin.exe", "")
        tools.save("source_frameworks/sframe1/include/include.h", "")
        tools.save("source_frameworks/sframe2/include/include.h", "")
        tools.save("source_frameworks/sframe1/lib/libframework.lib", "")
        tools.save("source_frameworks/sframe2/lib/libframework.lib", "")
        tools.save("source_frameworks/sframe2/foo/bar.txt", "")

    def build(self):
        tools.save("build_sources/build_stuff.cpp", "")
        tools.save("build_sources/subdir/othersubdir/selective_stuff.cpp", "")
        tools.save("build_includes/include3.h", "")
        tools.save("build_includes/include4.hpp", "")
        tools.save("build_libs/blibone.a", "")
        tools.save("build_libs/blibtwo.a", "")
        tools.save("build_bins/build_bin.exe", "")
        tools.save("build_frameworks/bframe1/include/include.h", "")
        tools.save("build_frameworks/bframe2/include/include.h", "")


    def layout(self):

        self.folders.source = "my_source"
        self.folders.build = "my_build"
        self.folders.generators = "my_build/generators"

        # Package locations
        self.cpp.package.includedirs = ["my_includes"]
        self.cpp.package.srcdirs = ["my_sources"]
        self.cpp.package.bindirs = ["my_bins"]
        self.cpp.package.libdirs = ["my_libs"]
        self.cpp.package.frameworkdirs = ["my_frameworks"]

        # Source CPP INFO
        self.cpp.source.srcdirs = ["source_sources"]
        self.cpp.source.includedirs = ["source_includes", "source_includes2"]
        self.cpp.source.libdirs = ["source_libs"]
        self.cpp.source.bindirs = ["source_bins"]
        self.cpp.source.frameworkdirs = ["source_frameworks"]

        # Build CPP INFO
        self.cpp.build.srcdirs = ["build_sources",
                                              "build_sources/subdir/othersubdir"]
        self.cpp.build.includedirs = ["build_includes"]
        self.cpp.build.libdirs = ["build_libs"]
        self.cpp.build.bindirs = ["build_bins"]
        self.cpp.build.frameworkdirs = ["build_frameworks"]

    def package(self):
        # Discard include3.h from source
        packager = AutoPackager(self)
        packager.patterns.build.include = ["*.hpp", "*.h", "include3.h"]
        packager.patterns.build.lib = ["*.a"]
        packager.patterns.build.bin = ["*.exe"]
        packager.patterns.build.src = ["*.cpp"]
        packager.patterns.build.framework = ["sframe*", "bframe*"]
        packager.patterns.source.include = ["*.hpp"] # Discard include3.h from source
        packager.patterns.source.lib = ["*.a"]
        packager.patterns.source.bin = ["*.exe"]
        packager.patterns.source.src = ["*.cpp"]
        packager.patterns.source.framework = ["sframe*"]
        packager.run()
    """
    client.save({"conanfile.py": conan_file})
    client.run("create . lib/1.0@")
    package_id = re.search(r"lib/1.0:(\S+)", str(client.out)).group(1)
    ref = ConanFileReference.loads("lib/1.0@")
    pref = PackageReference(ref, package_id)
    p_folder = client.cache.package_layout(ref).package(pref)

    def p_path(path):
        return os.path.join(p_folder, path)

    # Check directories that shouldn't exist
    assert not os.path.exists(p_path("include"))
    assert not os.path.exists(p_path("bin"))
    assert not os.path.exists(p_path("lib"))

    # Check files source
    assert os.path.exists(p_path("my_sources"))
    assert os.path.exists(p_path("my_sources/source_stuff.cpp"))

    assert os.path.exists(p_path("my_includes/include1.hpp"))
    assert os.path.exists(p_path("my_includes/include2.hpp"))
    assert not os.path.exists(p_path("my_includes/include.h"))

    assert os.path.exists(p_path("my_libs/slibone.a"))
    assert os.path.exists(p_path("my_libs/slibtwo.a"))

    assert os.path.exists(p_path("my_bins/source_bin.exe"))
    assert not os.path.exists(p_path("my_bins/bin_to_discard.exe"))

    assert os.path.exists(p_path("my_frameworks/sframe1/include/include.h"))
    assert os.path.exists(p_path("my_frameworks/sframe2/include/include.h"))
    assert os.path.exists(p_path("my_frameworks/sframe1/lib/libframework.lib"))
    assert os.path.exists(p_path("my_frameworks/sframe2/lib/libframework.lib"))
    assert os.path.exists(p_path("my_frameworks/sframe2/foo/bar.txt"))

    # Check files build
    assert os.path.exists(p_path("my_sources/build_stuff.cpp"))
    # note: as the "selective_stuff.cpp" is included in another my_sources dir, is copied twice
    #       it would need manual adjustement of the copy to avoid it
    assert os.path.exists(p_path("my_sources/selective_stuff.cpp"))
    assert os.path.exists(p_path("my_sources/subdir/othersubdir/selective_stuff.cpp"))

    assert os.path.exists(p_path("my_includes/include3.h"))
    assert not os.path.exists(p_path("my_includes/include4.h"))

    assert os.path.exists(p_path("my_libs/blibone.a"))
    assert os.path.exists(p_path("my_libs/blibtwo.a"))

    assert os.path.exists(p_path("my_bins/build_bin.exe"))

    assert os.path.exists(p_path("my_frameworks/sframe1/include/include.h"))
    assert os.path.exists(p_path("my_frameworks/sframe2/include/include.h"))


def test_auto_package_with_components():
    """The files from the components are not mixed in package if they belong to different dirs"""
    client = TestClient()
    conan_file = str(GenConanfile()
                     .with_settings("build_type")
                     .with_import("from conans import tools")
                     .with_import("from conan.tools.files import AutoPackager"))
    conan_file += """

    def source(self):
        tools.save("includes1/component1.hpp", "")
        tools.save("includes2/component2.hpp", "")

    def build(self):
        tools.save("build_libs/component1.a", "")
        tools.save("build_libs/component2.a", "")
        tools.save("build_bins/component3.exe", "")

    def layout(self):
        # Build and source infos
        self.cpp.source.components["component1"].includedirs = ["includes1"]
        self.cpp.source.components["component2"].includedirs = ["includes2"]
        self.cpp.build.components["component1"].libdirs = ["build_libs"]
        self.cpp.build.components["component2"].libdirs = ["build_libs"]
        self.cpp.build.components["component3"].bindirs = ["build_bins"]

        # Package infos
        self.cpp.package.components["component1"].includedirs = ["include"]
        self.cpp.package.components["component2"].includedirs = ["include2"]
        self.cpp.package.components["component1"].libdirs = ["lib"]
        self.cpp.package.components["component2"].libdirs = ["lib"]
        self.cpp.package.components["component3"].bindirs = ["bin"]

    def package(self):
        packager = AutoPackager(self)
        packager.run()
    """
    client.save({"conanfile.py": conan_file})
    client.run("create . lib/1.0@")
    package_id = re.search(r"lib/1.0:(\S+)", str(client.out)).group(1)
    ref = ConanFileReference.loads("lib/1.0@")
    pref = PackageReference(ref, package_id)
    p_folder = client.cache.package_layout(ref).package(pref)

    def p_path(path):
        return os.path.join(p_folder, path)

    # Check directories that shouldn't exist
    assert not os.path.exists(p_path("includes1"))
    assert not os.path.exists(p_path("includes2"))
    assert not os.path.exists(p_path("build_libs"))

    # Check includes
    assert os.path.exists(p_path("include"))
    assert os.path.exists(p_path("include/component1.hpp"))
    assert not os.path.exists(p_path("include/component2.hpp"))
    assert os.path.exists(p_path("include2/component2.hpp"))
    assert not os.path.exists(p_path("include2/component1.hpp"))

    # Check libs
    assert os.path.exists(p_path("lib/component1.a"))
    assert os.path.exists(p_path("lib/component2.a"))

    # Check app
    assert os.path.exists(p_path("bin/component3.exe"))


def test_auto_package_with_components_declared_badly():
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type")
                     .with_import("from conans import tools")
                     .with_import("from conan.tools.files import AutoPackager"))
    conan_file += """

    def layout(self):
        # Build and source infos
        self.cpp.source.components["component1"].includedirs = ["includes1"]
        self.cpp.source.components["component2"].includedirs = ["includes2"]
        self.cpp.build.components["component1"].libdirs = ["build_libs"]
        self.cpp.build.components["component2"].libdirs = ["build_libs"]
        self.cpp.build.components["component3"].bindirs = ["build_bins"]

        # Package infos BUT NOT DECLARING component2
        self.cpp.package.components["component1"].includedirs = ["include"]
        self.cpp.package.components["component1"].libdirs = ["lib"]
        self.cpp.package.components["component3"].bindirs = ["bin"]

    def package(self):
        AutoPackager(self).run()
    """

    client.save({"conanfile.py": conan_file})
    client.run("create . lib/1.0@", assert_error=True)
    assert "There are components declared in cpp.source.components or in " \
           "cpp.build.components that are not declared in " \
           "cpp.package.components" in client.out


def test_auto_package_default_patterns():
    """By default:
        self.patterns.source.include = ["*.h", "*.hpp", "*.hxx"]
        self.patterns.build.lib = ["*.so", "*.so.*", "*.a", "*.lib", "*.dylib"]
        self.patterns.build.bin = ["*.exe", "*.dll"]
    """
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type")
                     .with_import("from conans import tools")
                     .with_import("import os")
                     .with_import("from conan.tools.files import AutoPackager"))
    conan_file += """
    def source(self):
        tools.save("myincludes/mylib.header","")
        tools.save("myincludes/mylib.h","")
        tools.save("myincludes/mylib.hpp","")
        tools.save("myincludes/mylib.hxx","")

    def build(self):
        tools.save("ugly_build/mylib.a", "")
        tools.save("ugly_build/mylib.so", "")
        tools.save("ugly_build/mylib.so.0", "")
        tools.save("ugly_build/mylib.lib", "")
        tools.save("ugly_build/mylib.dylib", "")
        tools.save("ugly_build/app.exe", "")
        tools.save("ugly_build/app.dll", "")
        tools.save("ugly_build/mylib.janderclander", "")

    def layout(self):
        self.cpp.source.includedirs = ["myincludes"]
        self.cpp.build.libdirs = ["ugly_build"]
        self.cpp.build.bindirs = ["ugly_build"]

    def package(self):
        AutoPackager(self).run()
    """
    client.save({"conanfile.py": conan_file})
    client.run("create . lib/1.0@")
    package_id = re.search(r"lib/1.0:(\S+)", str(client.out)).group(1)
    ref = ConanFileReference.loads("lib/1.0@")
    pref = PackageReference(ref, package_id)
    p_folder = client.cache.package_layout(ref).package(pref)

    assert set(os.listdir(os.path.join(p_folder, "lib"))) == {"mylib.a", "mylib.so", "mylib.so.0",
                                                              "mylib.dylib", "mylib.lib"}
    assert set(os.listdir(os.path.join(p_folder, "include"))) == {"mylib.h", "mylib.hpp",
                                                                  "mylib.hxx"}
    assert set(os.listdir(os.path.join(p_folder, "bin"))) == {"app.exe", "app.dll"}


def test_auto_package_default_folders_with_components():
    """By default, the cpp_info of the components are empty"""
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type")
                     .with_import("from conans import tools")
                     .with_import("import os")
                     .with_import("from conan.tools.files import AutoPackager"))
    conan_file += """
    def layout(self):
        for el in [self.cpp.source, self.cpp.build]:
            # The defaults for cpp.build and cpp.source are empty
            assert el.components["foo"].includedirs is None
            assert el.components["foo"].libdirs is None
            assert el.components["foo"].bindirs is None
            assert el.components["foo"].frameworkdirs is None
            assert el.components["foo"].srcdirs is None
            assert el.components["foo"].resdirs is None

        # The defaults for cpp.package are filled includedirs and libdirs and bindirs
        assert self.cpp.package.components["foo"].includedirs is not None
        assert self.cpp.package.components["foo"].libdirs is not None
        assert self.cpp.package.components["foo"].bindirs is not None
        assert self.cpp.package.components["foo"].frameworkdirs is None
        assert self.cpp.package.components["foo"].srcdirs is None
        assert self.cpp.package.components["foo"].resdirs is None

    def package(self):
        AutoPackager(self).run()
    """
    client.save({"conanfile.py": conan_file})
    client.run("create . lib/1.0@")


def test_auto_package_with_custom_package_too():
    """We can also declare the package() method and call explicitly to the
       self.folders.package_files()"""
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type")
                     .with_import("from conans import tools")
                     .with_import("import os")
                     .with_import("from conan.tools.files import AutoPackager"))
    conan_file += """
    def source(self):
        tools.save("myincludes/mylib.header","")

    def build(self):
        tools.save("ugly_build/mylib.a", "")

    def layout(self):
        self.cpp.source.includedirs = ["myincludes"]
        self.cpp.build.libdirs = ["ugly_build"]

    def package(self):
        packager = AutoPackager(self)
        packager.patterns.source.include = ["*.header"]
        packager.patterns.build.lib = ["*.a"]
        packager.run()
        assert os.path.exists(os.path.join(self.package_folder, "include", "mylib.header"))
        assert os.path.exists(os.path.join(self.package_folder, "lib", "mylib.a"))
        self.output.warn("Package method called!")

    """
    client.save({"conanfile.py": conan_file})
    client.run("create . lib/1.0@")
    assert "Package method called!" in client.out


def test_auto_package_only_one_destination():
    """If the layout declares more than one destination folder it fails, because it cannot guess
    where to put the artifacts (very weird situation a package with two include/)"""
    client = TestClient()
    conan_file = str(GenConanfile().with_settings("build_type")
                     .with_import("from conans import tools")
                     .with_import("import os")
                     .with_import("from conan.tools.files import AutoPackager"))
    conan_file += """
    def source(self):
        tools.save("myincludes/mylib.header","")

    def build(self):
       tools.save("ugly_build/mylib.a", "")

    def layout(self):
       self.cpp.source.includedirs = ["myincludes"]
       self.cpp.build.libdirs = ["ugly_build"]
       self.cpp.build.bindirs = ["ugly_build"]
       self.cpp.build.frameworkdirs = ["ugly_build"]
       self.cpp.build.srcdirs = ["ugly_build"]
       self.cpp.build.builddirs = ["ugly_build"]
       self.cpp.build.resdirs = ["ugly_build"]

       self.cpp.package.{} = ["folder1", "folder2"]

    def package(self):
        packager = AutoPackager(self)
        packager.patterns.source.include = ["*.header"]
        packager.patterns.build.lib = ["*.a"]
        packager.run()

    """
    for dirs in ["includedirs", "builddirs", "bindirs", "srcdirs", "frameworkdirs", "libdirs",
                 "resdirs"]:
        client.save({"conanfile.py": conan_file.format(dirs)})
        client.run("create . lib/1.0@", assert_error=True)
        assert "The package has more than 1 cpp_info.{}, " \
               "cannot package automatically".format(dirs) in client.out



# coding=utf-8
import os
import unittest

import six

from conans.errors import ConanException
from conans.model.build_info import CppInfo, DepsCppInfo, DepCppInfo
from conans.test.utils.test_files import temp_folder
from conans.util.files import save


class CppInfoComponentsTest(unittest.TestCase):

    def test_components_set(self):
        cpp_info = CppInfo("", "root_folder")
        cpp_info.components["liba"].name = "LIBA"
        cpp_info.components["libb"].includedirs.append("includewhat")
        cpp_info.components["libc"].libs.append("thelibc")
        self.assertListEqual(list(cpp_info.components.keys()), ["liba", "libb", "libc"])
        self.assertEqual(cpp_info.components["liba"].get_name("any"), "LIBA")
        self.assertListEqual(cpp_info.components["libb"].includedirs, ["include", "includewhat"])
        self.assertListEqual(cpp_info.components["libc"].libs, ["thelibc"])

    def test_no_components_inside_components(self):
        cpp_info = CppInfo("", "root_folder")
        cpp_info.components["liba"].name = "LIBA"
        with self.assertRaises(AttributeError):
            _ = cpp_info.components["libb"].components

    def test_deps_cpp_info_libs(self):
        deps_cpp_info = DepsCppInfo()

        dep1 = CppInfo("dep1", "root")
        dep1.components["liba"].libs.append("liba")
        dep1.components["libb"].libs.append("libb")
        deps_cpp_info.add("dep1", DepCppInfo(dep1))

        dep2 = CppInfo("dep2", "root")
        dep2.components["libc"].libs.append("libc")
        dep2.components["libd"].libs.append("libd")
        deps_cpp_info.add("dep2", DepCppInfo(dep2))

        dep3 = CppInfo("dep3", "root")
        dep3.libs.append("libdep3")
        deps_cpp_info.add("dep3", DepCppInfo(dep3))

        self.assertListEqual(["liba", "libb"], deps_cpp_info["dep1"].libs)
        self.assertListEqual(["libc", "libd"], deps_cpp_info["dep2"].libs)
        self.assertListEqual(["libdep3"], deps_cpp_info["dep3"].libs)
        self.assertListEqual(["liba", "libb", "libc", "libd", "libdep3"],
                             list(deps_cpp_info.libs))

    def test_deps_cpp_info_paths(self):
        deps_cpp_info = DepsCppInfo()

        folder1 = temp_folder()
        dep1 = CppInfo("dep1", folder1)
        os.mkdir(os.path.join(folder1, "include"))
        os.mkdir(os.path.join(folder1, "includea"))
        os.mkdir(os.path.join(folder1, "includeb"))
        dep1.components["liba"].includedirs.append("includea")
        dep1.components["libb"].includedirs.append("includeb")
        deps_cpp_info.add("dep1", DepCppInfo(dep1))

        folder2 = temp_folder()
        dep2 = CppInfo("dep2", folder2)
        os.mkdir(os.path.join(folder2, "include"))
        os.mkdir(os.path.join(folder2, "includec"))
        os.mkdir(os.path.join(folder2, "included"))
        dep2.components["libc"].includedirs.append("includec")
        dep2.components["libd"].includedirs.append("included")
        deps_cpp_info.add("dep2", DepCppInfo(dep2))

        self.assertListEqual([os.path.join(folder1, "include"), os.path.join(folder1, "includea"),
                              os.path.join(folder1, "includeb")],
                             list(deps_cpp_info["dep1"].include_paths))
        self.assertListEqual([os.path.join(folder2, "include"), os.path.join(folder2, "includec"),
                              os.path.join(folder2, "included")],
                             list(deps_cpp_info["dep2"].include_paths))
        self.assertListEqual([os.path.join(folder1, "include"), os.path.join(folder1, "includea"),
                              os.path.join(folder1, "includeb"), os.path.join(folder2, "include"),
                              os.path.join(folder2, "includec"), os.path.join(folder2, "included")],
                             list(deps_cpp_info.include_paths))

    def test_deps_cpp_info_libs_defines_flags(self):
        deps_cpp_info = DepsCppInfo()

        dep1 = CppInfo("dep1", "root")
        dep1.components["liba"].libs.append("liba")
        dep1.components["liba"].defines.append("DEFINEA")
        dep1.components["liba"].system_libs.append("sysa")
        dep1.components["liba"].cxxflags.append("cxxflaga")
        dep1.components["liba"].cflags.append("cflaga")
        dep1.components["liba"].sharedlinkflags.append("slinka")
        dep1.components["liba"].frameworks.append("frameworka")
        dep1.components["liba"].exelinkflags.append("elinka")
        dep1.components["libb"].libs.append("libb")
        dep1.components["libb"].defines.append("DEFINEB")
        dep1.components["libb"].system_libs.append("sysb")
        dep1.components["libb"].cxxflags.append("cxxflagb")
        dep1.components["libb"].cflags.append("cflagb")
        dep1.components["libb"].sharedlinkflags.append("slinkb")
        dep1.components["libb"].frameworks.append("frameworkb")
        dep1.components["libb"].exelinkflags.append("elinkb")
        deps_cpp_info.add("dep1", DepCppInfo(dep1))

        dep2 = CppInfo("dep2", "root")
        dep2.components["libc"].libs.append("libc")
        dep2.components["libd"].libs.append("libd")
        dep2.components["systemlib"].system_libs = ["systemlib"]
        dep2.components["libc"].cxxflags = ["cxxflagc"]
        dep2.components["libd"].cflags = ["cflagd"]
        dep2.components["libc"].sharedlinkflags = ["slinkc"]
        dep2.components["libd"].sharedlinkflags = ["slinkd"]
        deps_cpp_info.add("dep2", DepCppInfo(dep2))

        self.assertListEqual(["liba", "libb"], deps_cpp_info["dep1"].libs)
        self.assertListEqual(["libc", "libd"], deps_cpp_info["dep2"].libs)
        self.assertListEqual(["liba", "libb", "libc", "libd"], list(deps_cpp_info.libs))

        self.assertListEqual(["DEFINEA", "DEFINEB"], deps_cpp_info["dep1"].defines)
        self.assertListEqual(["DEFINEA", "DEFINEB"], list(deps_cpp_info.defines))

        self.assertListEqual(["sysa", "sysb"], deps_cpp_info["dep1"].system_libs)
        self.assertListEqual(["systemlib"], deps_cpp_info["dep2"].system_libs)
        self.assertListEqual(["sysa", "sysb", "systemlib"], list(deps_cpp_info.system_libs))

        self.assertListEqual(["cxxflaga", "cxxflagb"], deps_cpp_info["dep1"].cxxflags)
        self.assertListEqual(["cxxflagc"], deps_cpp_info["dep2"].cxxflags)
        self.assertListEqual(["cxxflagc", "cxxflaga", "cxxflagb"], list(deps_cpp_info.cxxflags))

        self.assertListEqual(["cflaga", "cflagb"], deps_cpp_info["dep1"].cflags)
        self.assertListEqual(["cflagd"], deps_cpp_info["dep2"].cflags)
        self.assertListEqual(["cflagd", "cflaga", "cflagb"], list(deps_cpp_info.cflags))

        self.assertListEqual(["slinka", "slinkb"], deps_cpp_info["dep1"].sharedlinkflags)
        self.assertListEqual(["slinkc", "slinkd"], deps_cpp_info["dep2"].sharedlinkflags)
        self.assertListEqual(["slinkc", "slinkd", "slinka", "slinkb"],
                             list(deps_cpp_info.sharedlinkflags))

        self.assertListEqual(["frameworka", "frameworkb"], deps_cpp_info["dep1"].frameworks)
        self.assertListEqual(["frameworka", "frameworkb"], list(deps_cpp_info.frameworks))

        self.assertListEqual(["elinka", "elinkb"], deps_cpp_info["dep1"].exelinkflags)
        self.assertListEqual([], deps_cpp_info["dep2"].exelinkflags)
        self.assertListEqual(["elinka", "elinkb"], list(deps_cpp_info.exelinkflags))

    def test_deps_cpp_info_libs_release_debug(self):
        deps_cpp_info = DepsCppInfo()

        dep1 = CppInfo("dep1", "root")
        dep1.components["liba"].libs.append("liba")
        with self.assertRaises(AttributeError):
            dep1.release.components["libb"].libs.append("libb")
        with self.assertRaises(AttributeError):
            dep1.debug.components["libb"].libs.append("libb_d")
        deps_cpp_info.add("dep1", DepCppInfo(dep1))

        dep2 = CppInfo("dep2", "root")
        dep2.release.libs.append("libdep2")
        dep2.debug.libs.append("libdep2_d")
        with self.assertRaises(AttributeError):
            dep2.components["libc"].release.libs.append("libc")
        with self.assertRaises(AttributeError):
            dep2.components["libc"].debug.libs.append("libc_d")
        dep2.components["libc"].libs.append("libc")
        dep2.components["libc"].libs.append("libc2")
        deps_cpp_info.add("dep2", DepCppInfo(dep2))

        self.assertListEqual(["liba"], deps_cpp_info["dep1"].libs)
        self.assertListEqual(["libc", "libc2"], deps_cpp_info["dep2"].libs)
        self.assertListEqual(["liba", "libc", "libc2"], list(deps_cpp_info.libs))

        self.assertListEqual([], deps_cpp_info["dep1"].release.libs)
        self.assertListEqual(["libdep2"], deps_cpp_info["dep2"].release.libs)
        self.assertListEqual(["libdep2"], list(deps_cpp_info.release.libs))

        self.assertListEqual([], deps_cpp_info["dep1"].debug.libs)
        self.assertListEqual(["libdep2_d"], deps_cpp_info["dep2"].debug.libs)
        self.assertListEqual(["libdep2_d"], list(deps_cpp_info.debug.libs))

    def test_cpp_info_link_order(self):

        def _assert_link_order(sorted_libs):
            """
            Assert that dependent libs of a component are always found later in the list
            """
            assert sorted_libs, "'sorted_libs' is empty"
            for num, lib in enumerate(sorted_libs):
                component_name = lib[-1]
                for dep in info.components[component_name].requires:
                    for comp_lib in info.components[dep].libs:
                        self.assertIn(comp_lib, sorted_libs[num:])

        info = CppInfo("dep1", "")
        info.components["6"].libs = ["lib6"]
        info.components["6"].requires = ["4", "5"]
        info.components["5"].libs = ["lib5"]
        info.components["5"].requires = ["2"]
        info.components["4"].libs = ["lib4"]
        info.components["4"].requires = ["1"]
        info.components["3"].libs = ["lib3"]
        info.components["3"].requires = ["1"]
        info.components["1"].libs = ["lib1"]
        info.components["1"].requires = ["2"]
        info.components["2"].libs = ["lib2"]
        info.components["2"].requires = []
        dep_cpp_info = DepCppInfo(info)
        _assert_link_order(dep_cpp_info.libs)
        self.assertEqual(["lib6", "lib5", "lib4", "lib3", "lib1", "lib2"], dep_cpp_info.libs)
        deps_cpp_info = DepsCppInfo()
        deps_cpp_info.add("dep1", dep_cpp_info)
        self.assertEqual(["lib6", "lib5", "lib4", "lib3", "lib1", "lib2"], list(deps_cpp_info.libs))

        info = CppInfo("dep2", "")
        info.components["K"].libs = ["libK"]
        info.components["K"].requires = ["G", "H"]
        info.components["J"].libs = ["libJ"]
        info.components["J"].requires = ["F"]
        info.components["G"].libs = ["libG"]
        info.components["G"].requires = ["F"]
        info.components["H"].libs = ["libH"]
        info.components["H"].requires = ["F", "E"]
        info.components["L"].libs = ["libL"]
        info.components["L"].requires = ["I"]
        info.components["F"].libs = ["libF"]
        info.components["F"].requires = ["C", "D"]
        info.components["I"].libs = ["libI"]
        info.components["I"].requires = ["E"]
        info.components["C"].libs = ["libC"]
        info.components["C"].requires = ["A"]
        info.components["D"].libs = ["libD"]
        info.components["D"].requires = ["A"]
        info.components["E"].libs = ["libE"]
        info.components["E"].requires = ["A", "B"]
        info.components["A"].libs = ["libA"]
        info.components["A"].requires = []
        info.components["B"].libs = ["libB"]
        info.components["B"].requires = []
        dep_cpp_info = DepCppInfo(info)
        _assert_link_order(dep_cpp_info.libs)
        self.assertEqual(["libK", "libJ", "libG", "libH", "libL", "libF", "libI", "libC", "libD",
                          "libE", "libA", "libB"], dep_cpp_info.libs)
        deps_cpp_info.add("dep2", dep_cpp_info)
        self.assertEqual(["lib6", "lib5", "lib4", "lib3", "lib1", "lib2", "libK", "libJ", "libG",
                          "libH", "libL", "libF", "libI", "libC", "libD", "libE", "libA", "libB"],
                         list(deps_cpp_info.libs))

    def test_cppinfo_inexistent_component_dep(self):
        info = CppInfo("", None)
        info.components["LIB1"].requires = ["LIB2"]
        with six.assertRaisesRegex(self, ConanException, "Component 'LIB1' required components "
                                                         "not found in this package: 'LIB2'"):
            _ = DepCppInfo(info).libs
        info.components["LIB1"].requires = ["::LIB2"]
        with six.assertRaisesRegex(self, ConanException, "Leading character '::' not allowed in "
                                                         "LIB1 requires"):
            _ = DepCppInfo(info).libs

    def test_cpp_info_components_requires_loop(self):
        info = CppInfo("", "")
        info.components["LIB1"].requires = ["LIB1"]
        msg = "There is a dependency loop in 'self.cpp_info.components' requires"
        with six.assertRaisesRegex(self, ConanException, msg):
            _ = DepCppInfo(info).libs
        info = CppInfo("", "")
        info.components["LIB1"].requires = ["LIB2"]
        info.components["LIB2"].requires = ["LIB1", "LIB2"]
        with six.assertRaisesRegex(self, ConanException, msg):
            _ = DepCppInfo(info).build_paths
        info = CppInfo("", "")
        info.components["LIB1"].requires = ["LIB2"]
        info.components["LIB2"].requires = ["LIB3"]
        info.components["LIB3"].requires = ["LIB1"]
        with six.assertRaisesRegex(self, ConanException, msg):
            _ = DepCppInfo(info).defines

    def test_components_libs_order(self):
        info = CppInfo("dep1", "")
        info.components["liba"].libs = ["liba"]
        info.components["libb"].libs = ["libb"]
        dep_cpp_info = DepCppInfo(info)
        self.assertListEqual(["liba", "libb"], dep_cpp_info.libs)
        deps_cpp_info = DepsCppInfo()
        deps_cpp_info.add("dep1", dep_cpp_info)
        self.assertListEqual(["liba", "libb"], deps_cpp_info["dep1"].libs)
        self.assertListEqual(["liba", "libb"], list(deps_cpp_info.libs))

        info = CppInfo("dep1", "")
        info.components["liba"].libs = ["liba"]
        info.components["libb"].libs = ["libb"]
        dep_cpp_info = DepCppInfo(info)
        info2 = CppInfo("dep2", "")
        info2.components["libc"].libs = ["libc"]
        dep_cpp_info2 = DepCppInfo(info2)
        deps_cpp_info = DepsCppInfo()
        # Update in reverse order
        deps_cpp_info.add("dep2", dep_cpp_info2)
        deps_cpp_info.add("dep1", dep_cpp_info)
        self.assertListEqual(["liba", "libb"], deps_cpp_info["dep1"].libs)
        self.assertListEqual(["libc"], deps_cpp_info["dep2"].libs)
        self.assertListEqual(["libc", "liba", "libb"], list(deps_cpp_info.libs))

        info = CppInfo("dep1", "")
        info.components["liba"].libs = ["liba"]
        info.components["libb"].libs = ["libb"]
        info.components["libb"].requires = ["liba"]
        dep_cpp_info = DepCppInfo(info)
        self.assertListEqual(["libb", "liba"], dep_cpp_info.libs)
        deps_cpp_info = DepsCppInfo()
        deps_cpp_info.add("dep1", dep_cpp_info)
        self.assertListEqual(["libb", "liba"], deps_cpp_info["dep1"].libs)
        self.assertListEqual(["libb", "liba"], list(deps_cpp_info.libs))

        info = CppInfo("dep1", "")
        info.components["liba"].libs = ["liba"]
        info.components["libb"].libs = ["libb"]
        info.components["libb"].requires = ["liba"]
        dep_cpp_info = DepCppInfo(info)
        info2 = CppInfo("dep2", "")
        info2.components["libc"].libs = ["libc"]
        dep_cpp_info2 = DepCppInfo(info2)
        deps_cpp_info = DepsCppInfo()
        # Update in reverse order
        deps_cpp_info.add("dep2", dep_cpp_info2)
        deps_cpp_info.add("dep1", dep_cpp_info)
        self.assertListEqual(["libb", "liba"], deps_cpp_info["dep1"].libs)
        self.assertListEqual(["libc"], deps_cpp_info["dep2"].libs)
        self.assertListEqual(["libc", "libb", "liba"], list(deps_cpp_info.libs))

    def test_cppinfo_components_dirs(self):
        folder = temp_folder()
        info = CppInfo("OpenSSL", folder)
        info.components["OpenSSL"].includedirs = ["include"]
        info.components["OpenSSL"].libdirs = ["lib"]
        info.components["OpenSSL"].builddirs = ["build"]
        info.components["OpenSSL"].bindirs = ["bin"]
        info.components["OpenSSL"].resdirs = ["res"]
        info.components["Crypto"].includedirs = ["headers"]
        info.components["Crypto"].libdirs = ["libraries"]
        info.components["Crypto"].builddirs = ["build_scripts"]
        info.components["Crypto"].bindirs = ["binaries"]
        info.components["Crypto"].resdirs = ["resources"]
        self.assertEqual(["include"], info.components["OpenSSL"].includedirs)
        self.assertEqual(["lib"], info.components["OpenSSL"].libdirs)
        self.assertEqual(["build"], info.components["OpenSSL"].builddirs)
        self.assertEqual(["bin"], info.components["OpenSSL"].bindirs)
        self.assertEqual(["res"], info.components["OpenSSL"].resdirs)
        self.assertEqual(["headers"], info.components["Crypto"].includedirs)
        self.assertEqual(["libraries"], info.components["Crypto"].libdirs)
        self.assertEqual(["build_scripts"], info.components["Crypto"].builddirs)
        self.assertEqual(["binaries"], info.components["Crypto"].bindirs)
        self.assertEqual(["resources"], info.components["Crypto"].resdirs)

        info.components["Crypto"].includedirs = ["different_include"]
        info.components["Crypto"].libdirs = ["different_lib"]
        info.components["Crypto"].builddirs = ["different_build"]
        info.components["Crypto"].bindirs = ["different_bin"]
        info.components["Crypto"].resdirs = ["different_res"]
        self.assertEqual(["different_include"], info.components["Crypto"].includedirs)
        self.assertEqual(["different_lib"], info.components["Crypto"].libdirs)
        self.assertEqual(["different_build"], info.components["Crypto"].builddirs)
        self.assertEqual(["different_bin"], info.components["Crypto"].bindirs)
        self.assertEqual(["different_res"], info.components["Crypto"].resdirs)

        info.components["Crypto"].includedirs.extend(["another_include"])
        info.components["Crypto"].includedirs.append("another_other_include")
        info.components["Crypto"].libdirs.extend(["another_lib"])
        info.components["Crypto"].libdirs.append("another_other_lib")
        info.components["Crypto"].builddirs.extend(["another_build"])
        info.components["Crypto"].builddirs.append("another_other_build")
        info.components["Crypto"].bindirs.extend(["another_bin"])
        info.components["Crypto"].bindirs.append("another_other_bin")
        info.components["Crypto"].resdirs.extend(["another_res"])
        info.components["Crypto"].resdirs.append("another_other_res")
        self.assertEqual(["different_include", "another_include", "another_other_include"],
                         info.components["Crypto"].includedirs)
        self.assertEqual(["different_lib", "another_lib", "another_other_lib"],
                         info.components["Crypto"].libdirs)
        self.assertEqual(["different_build", "another_build", "another_other_build"],
                         info.components["Crypto"].builddirs)
        self.assertEqual(["different_bin", "another_bin", "another_other_bin"],
                         info.components["Crypto"].bindirs)
        self.assertEqual(["different_res", "another_res", "another_other_res"],
                         info.components["Crypto"].resdirs)

    def test_component_default_dirs_deps_cpp_info(self):
        folder = temp_folder()
        info = CppInfo("my_lib", folder)
        info.components["Component"].filter_empty = False  # For testing purposes
        dep_info = DepCppInfo(info)
        deps_cpp_info = DepsCppInfo()
        deps_cpp_info.add("my_lib", dep_info)
        self.assertListEqual([os.path.join(folder, "include")], list(deps_cpp_info.includedirs))
        self.assertListEqual([], list(deps_cpp_info.srcdirs))
        self.assertListEqual([os.path.join(folder, "lib")], list(deps_cpp_info.libdirs))
        self.assertListEqual([os.path.join(folder, "bin")], list(deps_cpp_info.bindirs))
        self.assertListEqual([os.path.join(folder, "")], list(deps_cpp_info.builddirs))
        self.assertListEqual([os.path.join(folder, "res")], list(deps_cpp_info.resdirs))
        self.assertListEqual([os.path.join(folder, "Frameworks")], list(deps_cpp_info.frameworkdirs))

    def test_deps_cpp_info_components(self):
        folder = temp_folder()
        info = CppInfo("my_lib", folder)
        # Create file so path is not cleared
        save(os.path.join(folder, "include", "my_file.h"), "")
        info.components["Component"].libs = ["libcomp"]
        dep_info = DepCppInfo(info)
        deps_cpp_info = DepsCppInfo()
        deps_cpp_info.add("my_lib", dep_info)
        self.assertListEqual(["libcomp"], list(deps_cpp_info.libs))
        self.assertListEqual(["libcomp"], deps_cpp_info["my_lib"].components["Component"].libs)
        self.assertListEqual([os.path.join(folder, "include")], list(deps_cpp_info.include_paths))
        self.assertListEqual([os.path.join(folder, "include")],
                             list(deps_cpp_info["my_lib"].components["Component"].include_paths))

    def test_deps_cpp_info_components_includedirs(self):
        info = CppInfo("my_lib", "root")
        info.components["component"].includedirs = ["include1", "include2"]
        info.components["component"].filter_empty = False
        dep_info = DepCppInfo(info)
        expected = [os.path.join("root", "include1"), os.path.join("root", "include2")]
        self.assertListEqual(expected, list(dep_info.include_paths))
        deps_cpp_info = DepsCppInfo()
        deps_cpp_info.add("my_lib", dep_info)
        self.assertListEqual(expected, list(deps_cpp_info.includedirs))



import unittest

from conans.test.utils.tools import GenConanfile, TestClient


class DownloadParallelTest(unittest.TestCase):

    def test_basic_parallel_download(self):
        client = TestClient(default_server_user=True)
        threads = 1  # At the moment, not really parallel until output implements mutex
        counter = 4
        client.run("config set general.parallel_download=%s" % threads)
        client.save({"conanfile.py": GenConanfile().with_option("myoption", '"ANY"')})

        for i in range(counter):
            client.run("create . pkg/0.1@user/testing -o pkg:myoption=%s" % i)
        client.run("upload * --all --confirm")
        client.run("remove * -f")

        # Lets download the packages
        client.run("download pkg/0.1@user/testing")
        self.assertIn("Downloading binary packages in %s parallel threads" % threads, client.out)
        self.assertIn("pkg/0.1@user/testing: Package installed "
                      "74ca4e392408c388db596b086fca5ebf64d825c0", client.out)
        self.assertIn("pkg/0.1@user/testing: Package installed "
                      "522dbc702b9cc2b582607ad6525f32ebd1442be5", client.out)
        self.assertIn("pkg/0.1@user/testing: Package installed "
                      "11997e24a862625b5e4753858f71aaf81a58a9b4", client.out)
        self.assertIn("pkg/0.1@user/testing: Package installed "
                      "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", client.out)



import textwrap

from conans.test.utils.conan_v2_tests import ConanV2ModeTestCase


class MesonBuildHelperTestCase(ConanV2ModeTestCase):
    profile = textwrap.dedent("""
        [settings]
        os = Linux
        arch = x86_64
        build_type = Release
        compiler=gcc
        compiler.version=4.9
        compiler.libcxx=libstdc++
        """)

    def test_no_build_type(self):
        t = self.get_client()
        conanfile = textwrap.dedent("""
            from conans import ConanFile, Meson
            class Pkg(ConanFile):
                settings = "os", "arch", "compiler"
                def build(self):
                     meson = Meson(self)
                     meson.build()
        """)
        t.save({"conanfile.py": conanfile, "myprofile": self.profile})
        t.run("create . pkg/0.1@user/testing -pr myprofile", assert_error=True)
        self.assertIn("Conan v2 incompatible: build_type setting should be defined.", t.out)

    def test_no_compiler(self):
        t = self.get_client()
        conanfile = textwrap.dedent("""
            from conans import ConanFile, Meson

            class Pkg(ConanFile):
                settings = "os", "arch", "build_type"
                def build(self):
                     meson = Meson(self)
                     meson.build()
        """)
        t.save({"conanfile.py": conanfile, "myprofile": self.profile})
        t.run("create . pkg/0.1@user/testing -pr myprofile", assert_error=True)
        self.assertIn("Conan v2 incompatible: compiler setting should be defined.", t.out)



import os
import platform

import pytest

from conan.tools.env.environment import environment_wrap_command
from conans.test.assets.pkg_cmake import pkg_cmake, pkg_cmake_app
from conans.test.assets.sources import gen_function_cpp
from conans.test.utils.test_files import temp_folder
from conans.test.utils.tools import TestClient


def editable_cmake(generator, build_folder=None):
    c = TestClient()
    if generator is not None:
        c.save({"global.conf": "tools.cmake.cmaketoolchain:generator={}".format(generator)},
               path=os.path.join(c.cache.cache_folder))
    c.save(pkg_cmake("dep", "0.1"), path=os.path.join(c.current_folder, "dep"))
    c.save(pkg_cmake_app("pkg", "0.1", requires=["dep/0.1"]),
           path=os.path.join(c.current_folder, "pkg"))

    output_folder = '--output-folder="{}"'.format(build_folder) if build_folder else ""
    build_folder = '--build-folder="{}"'.format(build_folder) if build_folder else ""

    def build_dep():
        c.run("install . -if=install_release {}".format(output_folder))
        c.run("build . -if=install_release {}".format(build_folder))
        c.run("install . -s build_type=Debug -if=install_debug {}".format(output_folder))
        c.run("build . -if=install_debug {}".format(build_folder))

    with c.chdir("dep"):
        c.run("editable add . dep/0.1@ {}".format(output_folder))
        build_dep()

    def build_pkg(msg):
        c.run("build . -if=install_release")
        folder = os.path.join("build", "Release")
        c.run_command(os.sep.join([".", folder, "pkg"]))
        assert "main: Release!" in c.out
        assert "{}: Release!".format(msg) in c.out
        c.run("build . -if=install_debug")
        folder = os.path.join("build", "Debug")
        c.run_command(os.sep.join([".", folder, "pkg"]))
        assert "main: Debug!" in c.out
        assert "{}: Debug!".format(msg) in c.out

    with c.chdir("pkg"):
        c.run("install . -if=install_release")
        c.run("install . -if=install_debug -s build_type=Debug")
        build_pkg("dep")

    # Do a source change in the editable!
    with c.chdir("dep"):
        c.save({"src/dep.cpp": gen_function_cpp(name="dep", msg="SUPERDEP")})
        build_dep()

    with c.chdir("pkg"):
        build_pkg("SUPERDEP")

    # Check that create is still possible
    c.run("editable remove dep/0.1@")
    c.run("create dep")
    c.run("create pkg")
    # print(c.out)
    assert "pkg/0.1: Created package" in c.out


@pytest.mark.skipif(platform.system() != "Windows", reason="Only windows")
@pytest.mark.parametrize("generator", [None, "MinGW Makefiles"])
@pytest.mark.tool_mingw64
def test_editable_cmake_windows(generator):
    editable_cmake(generator)


def test_editable_cmake_windows_folders():
    build_folder = temp_folder()
    editable_cmake(generator=None, build_folder=build_folder)


@pytest.mark.skipif(platform.system() != "Linux", reason="Only linux")
@pytest.mark.parametrize("generator", [None, "Ninja", "Ninja Multi-Config"])
def test_editable_cmake_linux(generator):
    editable_cmake(generator)


@pytest.mark.skipif(platform.system() != "Darwin", reason="Requires Macos")
@pytest.mark.parametrize("generator", [None, "Ninja", "Xcode"])
@pytest.mark.tool_cmake(version="3.19")
def test_editable_cmake_osx(generator):
    editable_cmake(generator)


def editable_cmake_exe(generator):
    # This test works because it is not multi-config or single config, but explicit in
    # --install folder
    c = TestClient()
    if generator is not None:
        c.save({"global.conf": "tools.cmake.cmaketoolchain:generator={}".format(generator)},
               path=os.path.join(c.cache.cache_folder))
    c.save(pkg_cmake("dep", "0.1", exe=True), path=os.path.join(c.current_folder, "dep"))

    def build_dep():
        c.run("install . -o dep:shared=True -if=install_release")
        c.run("build . -if=install_release")
        c.run("install . -s build_type=Debug -o dep:shared=True -if=install_debug")
        c.run("build . -if=install_debug")

    with c.chdir("dep"):
        c.run("editable add . dep/0.1@")
        build_dep()

    def run_pkg(msg):
        host_arch = c.get_default_host_profile().settings['arch']
        # FIXME: This only works with ``--install-folder``, layout() will break this
        cmd_release = environment_wrap_command(f"conanrunenv-release-{host_arch}", c.current_folder,
                                               "dep_app",)
        c.run_command(cmd_release)
        assert "{}: Release!".format(msg) in c.out
        cmd_release = environment_wrap_command(f"conanrunenv-debug-{host_arch}", c.current_folder,
                                               "dep_app", )
        c.run_command(cmd_release)
        assert "{}: Debug!".format(msg) in c.out

    with c.chdir("pkg"):
        c.run("install dep/0.1@ -o dep:shared=True -g VirtualRunEnv")
        c.run("install dep/0.1@ -o dep:shared=True -s build_type=Debug -g VirtualRunEnv")
        run_pkg("dep")

    # Do a source change in the editable!
    with c.chdir("dep"):
        c.save({"src/dep.cpp": gen_function_cpp(name="dep", msg="SUPERDEP")})
        build_dep()

    with c.chdir("pkg"):
        run_pkg("SUPERDEP")


@pytest.mark.skipif(platform.system() != "Windows", reason="Only windows")
@pytest.mark.parametrize("generator", [None, "MinGW Makefiles"])
@pytest.mark.tool_mingw64
def test_editable_cmake_windows_exe(generator):
    editable_cmake_exe(generator)


@pytest.mark.skipif(platform.system() != "Linux", reason="Only linux")
@pytest.mark.parametrize("generator", [None, "Ninja", "Ninja Multi-Config"])
def test_editable_cmake_linux_exe(generator):
    editable_cmake_exe(generator)


@pytest.mark.skipif(platform.system() != "Darwin", reason="Requires Macos")
@pytest.mark.parametrize("generator", [None, "Ninja", "Xcode"])
@pytest.mark.tool_cmake(version="3.19")
def test_editable_cmake_osx_exe(generator):
    editable_cmake_exe(generator)
